# BSD 2-Clause License
#
# Copyright (c) 2021-2024 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import typing as t

from ...entity.dbobject import DBModel, DBScript
from ...error import SSInternalError
from ..config import CONFIG
from ..utils.helpers import create_lockfile_name


def write_colocated_launch_script(
    file_name: str, db_log: str, colocated_settings: t.Dict[str, t.Any]
) -> None:
    """Write the colocated launch script

    This file will be written into the cwd of the step that
    is created for this entity.

    :param file_name: name of the script to write
    :param db_log: log file for the db
    :param colocated_settings: db settings from entity run_settings
    """

    colocated_cmd = _build_colocated_wrapper_cmd(db_log, **colocated_settings)

    with open(file_name, "w", encoding="utf-8") as script_file:
        script_file.write("#!/bin/bash\n")
        script_file.write("set -e\n\n")

        script_file.write("Cleanup () {\n")
        script_file.write("if ps -p $DBPID > /dev/null; then\n")
        script_file.write("\tkill -15 $DBPID\n")
        script_file.write("fi\n}\n\n")

        # run cleanup after all exitcodes
        script_file.write("trap Cleanup exit\n\n")

        # force entrypoint to write some debug information to the
        # STDOUT of the job
        if colocated_settings["debug"]:
            script_file.write("export SMARTSIM_LOG_LEVEL=debug\n")
        script_file.write(f"db_stdout=$({colocated_cmd})\n")
        # extract and set DBPID within the shell script that is
        # enclosed between __PID__ and sent to stdout by the colocated
        # entrypoints file
        script_file.write(
            "DBPID=$(echo $db_stdout | sed -n "
            "'s/.*__PID__\\([0-9]*\\)__PID__.*/\\1/p')\n"
        )

        # Write the actual launch command for the app
        script_file.write("$@\n\n")


def _build_colocated_wrapper_cmd(
    db_log: str,
    cpus: int = 1,
    rai_args: t.Optional[t.Dict[str, str]] = None,
    extra_db_args: t.Optional[t.Dict[str, str]] = None,
    port: int = 6780,
    ifname: t.Optional[t.Union[str, t.List[str]]] = None,
    custom_pinning: t.Optional[str] = None,
    **kwargs: t.Any,
) -> str:
    """Build the command use to run a colocated DB application

    :param db_log: log file for the db
    :param cpus: db cpus
    :param rai_args: redisai args
    :param extra_db_args: extra redis args
    :param port: port to bind DB to
    :param ifname: network interface(s) to bind DB to
    :param db_cpu_list: The list of CPUs that the database should be limited to
    :return: the command to run
    """
    # pylint: disable=too-many-locals

    # create unique lockfile name to avoid symlink vulnerability
    # this is the lockfile all the processes in the distributed
    # application will try to acquire. since we use a local tmp
    # directory on the compute node, only one process can acquire
    # the lock on the file.
    lockfile = create_lockfile_name()

    # create the command that will be used to launch the
    # database with the python entrypoint for starting
    # up the backgrounded db process

    cmd = [
        sys.executable,
        "-m",
        "smartsim._core.entrypoints.colocated",
        "+lockfile",
        lockfile,
        "+db_cpus",
        str(cpus),
    ]
    # Add in the interface if using TCP/IP
    if ifname:
        if isinstance(ifname, str):
            ifname = [ifname]
        cmd.extend(["+ifname", ",".join(ifname)])
    cmd.append("+command")
    # collect DB binaries and libraries from the config

    db_cmd = []
    if custom_pinning:
        db_cmd.extend(["taskset", "-c", custom_pinning])
    db_cmd.extend(
        [CONFIG.database_exe, CONFIG.database_conf, "--loadmodule", CONFIG.redisai]
    )

    # add extra redisAI configurations
    for arg, value in (rai_args or {}).items():
        if value:
            # RAI wants arguments for inference in all caps
            # ex. THREADS_PER_QUEUE=1
            db_cmd.append(f"{arg.upper()} {str(value)}")

    db_cmd.extend(["--port", str(port)])

    # Add socket and permissions for UDS
    unix_socket = kwargs.get("unix_socket", None)
    socket_permissions = kwargs.get("socket_permissions", None)

    if unix_socket and socket_permissions:
        db_cmd.extend(
            [
                "--unixsocket",
                str(unix_socket),
                "--unixsocketperm",
                str(socket_permissions),
            ]
        )
    elif bool(unix_socket) ^ bool(socket_permissions):
        raise SSInternalError(
            "`unix_socket` and `socket_permissions` must both be defined or undefined."
        )

    db_cmd.extend(
        ["--logfile", db_log]
    )  # usually /dev/null, unless debug was specified
    if extra_db_args:
        for db_arg, value in extra_db_args.items():
            # replace "_" with "-" in the db_arg because we use kwargs
            # for the extra configurations and Python doesn't allow a hyphen
            # in a variable name. All redis and KeyDB configuration options
            # use hyphens in their names.
            db_arg = db_arg.replace("_", "-")
            db_cmd.extend([f"--{db_arg}", value])

    db_models = kwargs.get("db_models", None)
    if db_models:
        db_model_cmd = _build_db_model_cmd(db_models)
        db_cmd.extend(db_model_cmd)

    db_scripts = kwargs.get("db_scripts", None)
    if db_scripts:
        db_script_cmd = _build_db_script_cmd(db_scripts)
        db_cmd.extend(db_script_cmd)

    cmd.extend(db_cmd)

    return " ".join(cmd)


def _build_db_model_cmd(db_models: t.List[DBModel]) -> t.List[str]:
    cmd = []
    for db_model in db_models:
        cmd.append("+db_model")
        cmd.append(f"--name={db_model.name}")

        # Here db_model.file is guaranteed to exist
        # because we don't allow the user to pass a serialized DBModel
        cmd.append(f"--file={db_model.file}")

        cmd.append(f"--backend={db_model.backend}")
        cmd.append(f"--device={db_model.device}")
        cmd.append(f"--devices_per_node={db_model.devices_per_node}")
        cmd.append(f"--first_device={db_model.first_device}")
        if db_model.batch_size:
            cmd.append(f"--batch_size={db_model.batch_size}")
        if db_model.min_batch_size:
            cmd.append(f"--min_batch_size={db_model.min_batch_size}")
        if db_model.min_batch_timeout:
            cmd.append(f"--min_batch_timeout={db_model.min_batch_timeout}")
        if db_model.tag:
            cmd.append(f"--tag={db_model.tag}")
        if db_model.inputs:
            cmd.append("--inputs=" + ",".join(db_model.inputs))
        if db_model.outputs:
            cmd.append("--outputs=" + ",".join(db_model.outputs))

    return cmd


def _build_db_script_cmd(db_scripts: t.List[DBScript]) -> t.List[str]:
    cmd = []
    for db_script in db_scripts:
        cmd.append("+db_script")
        cmd.append(f"--name={db_script.name}")
        if db_script.func:
            # Notice that here db_script.func is guaranteed to be a str
            # because we don't allow the user to pass a serialized function
            func = db_script.func
            sanitized_func = func.replace("\n", "\\n")
            if not (
                sanitized_func.startswith("'")
                and sanitized_func.endswith("'")
                or (sanitized_func.startswith('"') and sanitized_func.endswith('"'))
            ):
                sanitized_func = '"' + sanitized_func + '"'
            cmd.append(f"--func={sanitized_func}")
        elif db_script.file:
            cmd.append(f"--file={db_script.file}")
        cmd.append(f"--device={db_script.device}")
        cmd.append(f"--devices_per_node={db_script.devices_per_node}")
        cmd.append(f"--first_device={db_script.first_device}")
    return cmd



# BSD 2-Clause License
#
# Copyright (c) 2021-2024 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import abc
import asyncio
import collections
import itertools
import logging
import typing as t

import redis.asyncio as redisa
import redis.exceptions as redisex

from smartsim._core.control.job import JobEntity
from smartsim._core.utils.helpers import get_ts_ms
from smartsim._core.utils.telemetry.sink import FileSink, Sink

logger = logging.getLogger("TelemetryMonitor")


class Collector(abc.ABC):
    """Base class for telemetry collectors.

    A Collector is used to retrieve runtime metrics about an entity."""

    def __init__(self, entity: JobEntity, sink: Sink) -> None:
        """Initialize the collector

        :param entity: entity to collect metrics on
        :param sink: destination to write collected information
        """
        self._entity = entity
        self._sink = sink
        self._enabled = True

    @property
    def enabled(self) -> bool:
        """Boolean indicating if the collector should perform data collection"""
        return self._entity.telemetry_on

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._entity.telemetry_on = value

    @property
    def entity(self) -> JobEntity:
        """The `JobEntity` for which data is collected
        :return: the entity"""
        return self._entity

    @property
    def sink(self) -> Sink:
        """The sink where collected data is written
        :return: the sink
        """
        return self._sink

    @abc.abstractmethod
    async def prepare(self) -> None:
        """Initialization logic for the collector"""

    @abc.abstractmethod
    async def collect(self) -> None:
        """Execute metric collection"""

    @abc.abstractmethod
    async def shutdown(self) -> None:
        """Execute cleanup of resources for the collector"""


class _DBAddress:
    """Helper class to hold and pretty-print connection details"""

    def __init__(self, host: str, port: int) -> None:
        """Initialize the instance
        :param host: host address for database connections
        :param port: port number for database connections
        """
        self.host = host.strip() if host else ""
        self.port = port
        self._check()

    def _check(self) -> None:
        """Validate input arguments"""
        if not self.host:
            raise ValueError(f"{type(self).__name__} requires host")
        if not self.port:
            raise ValueError(f"{type(self).__name__} requires port")

    def __str__(self) -> str:
        """Pretty-print the instance"""
        return f"{self.host}:{self.port}"


class DBCollector(Collector):
    """A base class for collectors that retrieve statistics from an orchestrator"""

    def __init__(self, entity: JobEntity, sink: Sink) -> None:
        """Initialize the `DBCollector`

        :param entity: entity with metadata about the resource to monitor
        :param sink: destination to write collected information
        """
        super().__init__(entity, sink)
        self._client: t.Optional[redisa.Redis[bytes]] = None
        self._address = _DBAddress(
            self._entity.config.get("host", ""),
            int(self._entity.config.get("port", 0)),
        )

    async def _configure_client(self) -> None:
        """Configure the client connection to the target database"""
        try:
            if not self._client:
                self._client = redisa.Redis(
                    host=self._address.host, port=self._address.port
                )
        except Exception as e:
            logger.exception(e)
        finally:
            if not self._client:
                logger.error(
                    f"{type(self).__name__} failed to connect to {self._address}"
                )

    async def prepare(self) -> None:
        """Initialization logic for the DB collector. Creates a database
        connection then executes the `post_prepare` callback function."""
        if self._client:
            return

        await self._configure_client()
        await self._post_prepare()

    @abc.abstractmethod
    async def _post_prepare(self) -> None:
        """Hook function to enable subclasses to perform actions
        after a db client is ready"""

    @abc.abstractmethod
    async def _perform_collection(
        self,
    ) -> t.Sequence[t.Tuple[t.Union[int, float, str], ...]]:
        """Hook function for subclasses to execute custom metric retrieval.
        NOTE: all implementations return an iterable of metrics to avoid
        adding extraneous base class code to differentiate the results

        :return: an iterable containing individual metric collection results
        """

    async def collect(self) -> None:
        """Execute database metric collection if the collector is enabled. Writes
        the resulting metrics to the associated output sink. Calling `collect`
        when `self.enabled` is `False` performs no actions."""
        if not self.enabled:
            # collectors may be disabled by monitoring changes to the
            # manifest. Leave the collector but do NOT collect
            logger.debug(f"{type(self).__name__} is not enabled")
            return

        await self.prepare()
        if not self._client:
            logger.warning(f"{type(self).__name__} cannot collect")
            return

        try:
            # if we can't communicate w/the db, exit
            if not await self._check_db():
                return

            all_metrics = await self._perform_collection()
            for metrics in all_metrics:
                await self._sink.save(*metrics)
        except Exception as ex:
            logger.warning(f"Collect failed for {type(self).__name__}", exc_info=ex)

    async def shutdown(self) -> None:
        """Execute cleanup of database client connections"""
        try:
            if self._client:
                logger.info(
                    f"Shutting down {self._entity.name}::{self.__class__.__name__}"
                )
                await self._client.close()
                self._client = None
        except Exception as ex:
            logger.error(
                f"An error occurred during {type(self).__name__} shutdown", exc_info=ex
            )

    async def _check_db(self) -> bool:
        """Check if the target database is reachable.

        :return: `True` if connection succeeds, `False` otherwise.
        """
        try:
            if self._client:
                return await self._client.ping()
        except redisex.ConnectionError:
            logger.warning(f"Cannot ping db {self._address}")

        return False


class DBMemoryCollector(DBCollector):
    """A `DBCollector` that collects memory consumption metrics"""

    def __init__(self, entity: JobEntity, sink: Sink) -> None:
        super().__init__(entity, sink)
        self._columns = ["used_memory", "used_memory_peak", "total_system_memory"]

    async def _post_prepare(self) -> None:
        """Write column headers for a CSV formatted output sink after
        the database connection is established"""
        await self._sink.save("timestamp", *self._columns)

    async def _perform_collection(
        self,
    ) -> t.Sequence[t.Tuple[int, float, float, float]]:
        """Perform memory metric collection and return the results

        :return: an iterable containing individual metric collection results
        in the format `(timestamp,used_memory,used_memory_peak,total_system_memory)`
        """
        if self._client is None:
            return []

        db_info = await self._client.info("memory")

        used = float(db_info["used_memory"])
        peak = float(db_info["used_memory_peak"])
        total = float(db_info["total_system_memory"])

        value = (get_ts_ms(), used, peak, total)

        # return a list containing a single record to simplify the parent
        # class code to save multiple records from a single collection
        return [value]


class DBConnectionCollector(DBCollector):
    """A `DBCollector` that collects database client-connection metrics"""

    def __init__(self, entity: JobEntity, sink: Sink) -> None:
        super().__init__(entity, sink)
        self._columns = ["client_id", "address"]

    async def _post_prepare(self) -> None:
        """Write column headers for a CSV formatted output sink after
        the database connection is established"""
        await self._sink.save("timestamp", *self._columns)

    async def _perform_collection(
        self,
    ) -> t.Sequence[t.Tuple[t.Union[int, str, str], ...]]:
        """Perform connection metric collection and return the results

        :return: an iterable containing individual metric collection results
        in the format `(timestamp,client_id,address)`
        """
        if self._client is None:
            return []

        now_ts = get_ts_ms()
        clients = await self._client.client_list()

        values: t.List[t.Tuple[int, str, str]] = []

        # content-filter the metrics and return them all together
        for client in clients:
            # all records for the request will have the same timestamp
            value = now_ts, client["id"], client["addr"]
            values.append(value)

        return values


class DBConnectionCountCollector(DBCollector):
    """A DBCollector that collects aggregated client-connection count metrics"""

    def __init__(self, entity: JobEntity, sink: Sink) -> None:
        super().__init__(entity, sink)
        self._columns = ["num_clients"]

    async def _post_prepare(self) -> None:
        """Write column headers for a CSV formatted output sink after
        the database connection is established"""
        await self._sink.save("timestamp", *self._columns)

    async def _perform_collection(
        self,
    ) -> t.Sequence[t.Tuple[int, int]]:
        """Perform connection-count metric collection and return the results

        :return: an iterable containing individual metric collection results
        in the format `(timestamp,num_clients)`
        """
        if self._client is None:
            return []

        client_list = await self._client.client_list()

        addresses = {item["addr"] for item in client_list}

        # return a list containing a single record to simplify the parent
        # class code to save multiple records from a single collection
        value = (get_ts_ms(), len(addresses))
        return [value]


class CollectorManager:
    """The `CollectorManager` manages the set of all collectors required to retrieve
    metrics for an experiment. It provides the ability to add and remove collectors
    with unique configuration per entity. The `CollectorManager` is primarily used
    to perform bulk actions on 1-to-many collectors (e.g. prepare all collectors,
    request metrics for all collectors, close all collector connections)"""

    def __init__(self, timeout_ms: int = 1000) -> None:
        """Initialize the `CollectorManager` without collectors
        :param timeout_ms: maximum time (in ms) allowed for `Collector.collect`
        """
        # A lookup table to hold a list of registered collectors per entity
        self._collectors: t.Dict[str, t.List[Collector]] = collections.defaultdict(list)
        # Max time to allow a collector to work before cancelling requests
        self._timeout_ms = timeout_ms

    def clear(self) -> None:
        """Remove all collectors from the monitored set"""
        self._collectors = collections.defaultdict(list)

    def add(self, collector: Collector) -> None:
        """Add a collector to the monitored set

        :param collector: `Collector` instance to monitor
        """
        entity_name = collector.entity.name

        registered_collectors = self._collectors[entity_name]

        # Exit if the collector is already registered to the entity
        if any(c for c in registered_collectors if type(c) is type(collector)):
            return

        logger.debug(f"Adding collector: {entity_name}::{type(collector).__name__}")
        registered_collectors.append(collector)

    def add_all(self, collectors: t.Sequence[Collector]) -> None:
        """Add multiple collectors to the monitored set

        :param collectors: a collection of `Collectors` to monitor
        """
        for collector in collectors:
            self.add(collector)

    async def remove_all(self, entities: t.Sequence[JobEntity]) -> None:
        """Remove all collectors registered to the supplied entities

        :param entities: a collection of `JobEntity` instances that will
        no longer have registered collectors
        """
        if not entities:
            return

        tasks = (self.remove(entity) for entity in entities)
        await asyncio.gather(*tasks)

    async def remove(self, entity: JobEntity) -> None:
        """Remove all collectors registered to the supplied entity

        :param entities: `JobEntity` that will no longer have registered collectors
        """
        registered = self._collectors.pop(entity.name, [])
        if not registered:
            return

        logger.debug(f"Removing collectors registered for {entity.name}")
        asyncio.gather(*(collector.shutdown() for collector in registered))

    async def prepare(self) -> None:
        """Prepare registered collectors to perform collection"""
        tasks = (collector.prepare() for collector in self.all_collectors)
        # use gather so all collectors are prepared before collection
        await asyncio.gather(*tasks)

    async def collect(self) -> None:
        """Perform collection for all registered collectors"""
        if collectors := self.all_collectors:
            tasks = [asyncio.create_task(item.collect()) for item in collectors]

            _, pending = await asyncio.wait(tasks, timeout=self._timeout_ms / 1000.0)

            # any tasks still pending has exceeded the timeout
            if pending:
                # manually cancel tasks since asyncio.wait will not
                for remaining_task in pending:
                    remaining_task.cancel()
                logger.debug(f"Execution of {len(pending)} collectors timed out.")

    async def shutdown(self) -> None:
        """Release resources for all registered collectors"""
        logger.debug(f"{type(self).__name__} shutting down collectors...")
        if list(self.all_collectors):
            shutdown_tasks = []
            # create an async tasks to execute all shutdowns in parallel
            for item in self.all_collectors:
                shutdown_tasks.append(asyncio.create_task(item.shutdown()))
            # await until all shutdowns are complete
            await asyncio.wait(shutdown_tasks)
        logger.debug("Collector shutdown complete...")

    @property
    def all_collectors(self) -> t.Sequence[Collector]:
        """Get a list of all registered collectors

        :return: a collection of registered collectors for all entities
        """
        # flatten and return all the lists-of-collectors that are registered
        collectors = itertools.chain.from_iterable(self._collectors.values())
        return [collector for collector in collectors if collector.enabled]

    @property
    def dead_collectors(self) -> t.Sequence[Collector]:
        """Get a list of all disabled collectors

        :return: a collection of disabled collectors for all entities
        """
        collectors = itertools.chain.from_iterable(self._collectors.values())
        return [collector for collector in collectors if not collector.enabled]

    def register_collectors(self, entity: JobEntity) -> None:
        """Find all configured collectors for the entity and register them

        :param entity: a `JobEntity` instance that will have all configured collectors
        registered for collection. Configuration is found in the `RuntimeManifest`
        """
        collectors: t.List[Collector] = []

        # ONLY db telemetry is implemented at this time. This resolver must
        # be updated when non-database or always-on collectors are introduced
        if entity.is_db and entity.telemetry_on:
            if mem_out := entity.collectors.get("memory", None):
                collectors.append(DBMemoryCollector(entity, FileSink(mem_out)))

            if con_out := entity.collectors.get("client", None):
                collectors.append(DBConnectionCollector(entity, FileSink(con_out)))

            if num_out := entity.collectors.get("client_count", None):
                collectors.append(DBConnectionCountCollector(entity, FileSink(num_out)))
        else:
            logger.debug(f"Collectors disabled for db {entity.name}")

        self.add_all(collectors)

    def register_all_collectors(self, entities: t.Sequence[JobEntity]) -> None:
        """Find all configured collectors for the entity and register them

        :param entities: entities to call `register_collectors` for
        """
        for entity in entities:
            self.register_collectors(entity)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import copy
import typing as t
from pprint import pformat

from ..error import SSUnsupportedError
from ..log import get_logger
from .base import BatchSettings, RunSettings

logger = get_logger(__name__)


class JsrunSettings(RunSettings):
    def __init__(
        self,
        exe: str,
        exe_args: t.Optional[t.Union[str, t.List[str]]] = None,
        run_args: t.Optional[t.Dict[str, t.Union[int, str, float, None]]] = None,
        env_vars: t.Optional[t.Dict[str, t.Optional[str]]] = None,
        **_kwargs: t.Any,
    ) -> None:
        """Settings to run job with ``jsrun`` command

        ``JsrunSettings`` should only be used on LSF-based systems.

        :param exe: executable
        :param exe_args: executable arguments
        :param run_args: arguments for run command
        :param env_vars: environment vars to launch job with
        """
        super().__init__(
            exe,
            exe_args,
            run_command="jsrun",
            run_args=run_args,
            env_vars=env_vars,
        )

        # Parameters needed for MPMD run
        self.erf_sets = {"host": "*", "cpu": "*", "ranks": "1"}
        self.mpmd_preamble_lines: t.List[str] = []
        self.mpmd: t.List[RunSettings] = []
        self.individual_suffix = ""

    reserved_run_args = {"chdir", "h"}

    def set_num_rs(self, num_rs: t.Union[str, int]) -> None:
        """Set the number of resource sets to use

        This sets ``--nrs``.

        :param num_rs: Number of resource sets or `ALL_HOSTS`
        """
        if isinstance(num_rs, str):
            self.run_args["nrs"] = num_rs
        else:
            self.run_args["nrs"] = int(num_rs)

    def set_cpus_per_rs(self, cpus_per_rs: int) -> None:
        """Set the number of cpus to use per resource set

        This sets ``--cpu_per_rs``

        :param cpus_per_rs: number of cpus to use per resource set or ALL_CPUS
        """
        if self.colocated_db_settings:
            db_cpus = int(t.cast(int, self.colocated_db_settings.get("db_cpus", 0)))
            if not db_cpus:
                raise ValueError("db_cpus must be configured on colocated_db_settings")

            if cpus_per_rs < db_cpus:
                raise ValueError(
                    f"Cannot set cpus_per_rs ({cpus_per_rs}) to less than "
                    + f"db_cpus ({db_cpus})"
                )
        if isinstance(cpus_per_rs, str):
            self.run_args["cpu_per_rs"] = cpus_per_rs
        else:
            self.run_args["cpu_per_rs"] = int(cpus_per_rs)

    def set_gpus_per_rs(self, gpus_per_rs: int) -> None:
        """Set the number of gpus to use per resource set

        This sets ``--gpu_per_rs``

        :param gpus_per_rs: number of gpus to use per resource set or ALL_GPUS
        """
        if isinstance(gpus_per_rs, str):
            self.run_args["gpu_per_rs"] = gpus_per_rs
        else:
            self.run_args["gpu_per_rs"] = int(gpus_per_rs)

    def set_rs_per_host(self, rs_per_host: int) -> None:
        """Set the number of resource sets to use per host

        This sets ``--rs_per_host``

        :param rs_per_host: number of resource sets to use per host
        """
        self.run_args["rs_per_host"] = int(rs_per_host)

    def set_tasks(self, tasks: int) -> None:
        """Set the number of tasks for this job

        This sets ``--np``

        :param tasks: number of tasks
        """
        self.run_args["np"] = int(tasks)

    def set_tasks_per_rs(self, tasks_per_rs: int) -> None:
        """Set the number of tasks per resource set

        This sets ``--tasks_per_rs``

        :param tasks_per_rs: number of tasks per resource set
        """
        self.run_args["tasks_per_rs"] = int(tasks_per_rs)

    def set_tasks_per_node(self, tasks_per_node: int) -> None:
        """Set the number of tasks per resource set.

        This function is an alias for `set_tasks_per_rs`.

        :param tasks_per_node: number of tasks per resource set
        """
        self.set_tasks_per_rs(int(tasks_per_node))

    def set_cpus_per_task(self, cpus_per_task: int) -> None:
        """Set the number of cpus per tasks.

        This function is an alias for `set_cpus_per_rs`.

        :param cpus_per_task: number of cpus per resource set
        """
        self.set_cpus_per_rs(int(cpus_per_task))

    def set_memory_per_rs(self, memory_per_rs: int) -> None:
        """Specify the number of megabytes of memory to assign to a resource set

        This sets ``--memory_per_rs``

        :param memory_per_rs: Number of megabytes per rs
        """
        self.run_args["memory_per_rs"] = int(memory_per_rs)

    def set_memory_per_node(self, memory_per_node: int) -> None:
        """Specify the number of megabytes of memory to assign to a resource set

        Alias for `set_memory_per_rs`.

        :param memory_per_node: Number of megabytes per rs
        """
        self.set_memory_per_rs(int(memory_per_node))

    def set_binding(self, binding: str) -> None:
        """Set binding

        This sets ``--bind``

        :param binding: Binding, e.g. `packed:21`
        """
        self.run_args["bind"] = binding

    def make_mpmd(self, settings: RunSettings) -> None:
        """Make step an MPMD (or SPMD) job.

        This method will activate job execution through an ERF file.

        Optionally, this method adds an instance of ``JsrunSettings`` to
        the list of settings to be launched in the same ERF file.

        :param settings: ``JsrunSettings`` instance
        """
        if self.colocated_db_settings:
            raise SSUnsupportedError(
                "Colocated models cannot be run as a mpmd workload"
            )

        self.mpmd.append(settings)

    def set_mpmd_preamble(self, preamble_lines: t.List[str]) -> None:
        """Set preamble used in ERF file. Typical lines include
        `oversubscribe-cpu : allow` or `overlapping-rs : allow`.
        Can be used to set `launch_distribution`. If it is not present,
        it will be inferred from the settings, or set to `packed` by
        default.

        :param preamble_lines: lines to put at the beginning of the ERF
                               file.
        """
        self.mpmd_preamble_lines = preamble_lines

    def set_erf_sets(self, erf_sets: t.Dict[str, str]) -> None:
        """Set resource sets used for ERF (SPMD or MPMD) steps.

        ``erf_sets`` is a dictionary used to fill the ERF
        line representing these settings, e.g.
        `{"host": "1", "cpu": "{0:21}, {21:21}", "gpu": "*"}`
        can be used to specify rank (or rank_count), hosts, cpus, gpus,
        and memory.
        The key `rank` is used to give specific ranks, as in
        `{"rank": "1, 2, 5"}`, while the key `rank_count` is used to specify
        the count only, as in `{"rank_count": "3"}`. If both are specified,
        only `rank` is used.

        :param hosts: dictionary of resources
        """
        self.erf_sets = copy.deepcopy(erf_sets)

    def format_env_vars(self) -> t.List[str]:
        """Format environment variables. Each variable needs
        to be passed with ``--env``. If a variable is set to ``None``,
        its value is propagated from the current environment.

        :returns: formatted list of strings to export variables
        """
        format_str = []
        for k, v in self.env_vars.items():
            if v:
                format_str += ["-E", f"{k}={v}"]
            else:
                format_str += ["-E", f"{k}"]
        return format_str

    def set_individual_output(self, suffix: t.Optional[str] = None) -> None:
        """Set individual std output.

        This sets ``--stdio_mode individual``
        and inserts the suffix into the output name. The resulting
        output name will be ``self.name + suffix + .out``.

        :param suffix: Optional suffix to add to output file names,
                       it can contain `%j`, `%h`, `%p`, or `%t`,
                       as specified by `jsrun` options.
        """
        self.run_args["stdio_mode"] = "individual"
        if suffix:
            self.individual_suffix = suffix

    def format_run_args(self) -> t.List[str]:
        """Return a list of LSF formatted run arguments

        :return: list of LSF arguments for these settings
        """
        # args launcher uses
        args = []
        restricted = ["chdir", "h", "stdio_stdout", "o", "stdio_stderr", "k"]
        if self.mpmd or "erf_input" in self.run_args.keys():
            restricted.extend(
                [
                    "tasks_per_rs",
                    "a",
                    "np",
                    "p",
                    "cpu_per_rs",
                    "c",
                    "gpu_per_rs",
                    "g",
                    "latency_priority",
                    "l",
                    "memory_per_rs",
                    "m",
                    "nrs",
                    "n",
                    "rs_per_host",
                    "r",
                    "rs_per_socket",
                    "K",
                    "appfile",
                    "f",
                    "allocate_only",
                    "A",
                    "launch_node_task",
                    "H",
                    "use_reservation",
                    "J",
                    "use_resources",
                    "bind",
                    "b",
                    "launch_distribution",
                    "d",
                ]
            )

        for opt, value in self.run_args.items():
            if opt not in restricted:
                short_arg = bool(len(str(opt)) == 1)
                prefix = "-" if short_arg else "--"
                if not value:
                    args += [prefix + opt]
                else:
                    if short_arg:
                        args += [prefix + opt, str(value)]
                    else:
                        args += ["=".join((prefix + opt, str(value)))]
        return args

    def __str__(self) -> str:
        string = super().__str__()
        if self.mpmd:
            string += "\nERF settings: " + pformat(self.erf_sets)
        return string

    def _prep_colocated_db(self, db_cpus: int) -> None:
        cpus_per_flag_set = False
        for cpu_per_rs_flag in ["cpu_per_rs", "c"]:
            if run_arg_value := self.run_args.get(cpu_per_rs_flag, 0):
                cpus_per_flag_set = True
                cpu_per_rs = int(run_arg_value)
                if cpu_per_rs < db_cpus:
                    msg = (
                        f"{cpu_per_rs_flag} flag was set to {cpu_per_rs}, but "
                        f"colocated DB requires {db_cpus} CPUs per RS. Automatically "
                        f"setting {cpu_per_rs_flag} flag to {db_cpus}"
                    )
                    logger.info(msg)
                    self.run_args[cpu_per_rs_flag] = db_cpus
        if not cpus_per_flag_set:
            msg = f"Colocated DB requires {db_cpus} CPUs per RS. Automatically setting "
            msg += f"--cpus_per_rs=={db_cpus}"
            logger.info(msg)
            self.set_cpus_per_rs(db_cpus)

        rs_per_host_set = False
        for rs_per_host_flag in ["rs_per_host", "r"]:
            if rs_per_host_flag in self.run_args:
                rs_per_host_set = True
                rs_per_host = self.run_args[rs_per_host_flag]
                if rs_per_host != 1:
                    msg = f"{rs_per_host_flag} flag was set to {rs_per_host}, "
                    msg += (
                        "but colocated DB requires running ONE resource set per host. "
                    )
                    msg += f"Automatically setting {rs_per_host_flag} flag to 1"
                    logger.info(msg)
                    self.run_args[rs_per_host_flag] = "1"
        if not rs_per_host_set:
            msg = "Colocated DB requires one resource set per host. "
            msg += " Automatically setting --rs_per_host==1"
            logger.info(msg)
            self.set_rs_per_host(1)


class BsubBatchSettings(BatchSettings):
    def __init__(
        self,
        nodes: t.Optional[int] = None,
        time: t.Optional[str] = None,
        project: t.Optional[str] = None,
        batch_args: t.Optional[t.Dict[str, t.Optional[str]]] = None,
        smts: int = 0,
        **kwargs: t.Any,
    ) -> None:
        """Specify ``bsub`` batch parameters for a job

        :param nodes: number of nodes for batch
        :param time: walltime for batch job in format hh:mm
        :param project: project for batch launch
        :param batch_args: overrides for LSF batch arguments
        :param smts: SMTs
        """
        self.project: t.Optional[str] = None

        if project:
            kwargs.pop("account", None)
        else:
            project = kwargs.pop("account", None)

        super().__init__(
            "bsub",
            batch_args=batch_args,
            nodes=nodes,
            account=project,
            time=time,
            **kwargs,
        )

        self.smts = 0
        if smts:
            self.set_smts(smts)

        self.expert_mode = False
        self.easy_settings = ["ln_slots", "ln_mem", "cn_cu", "nnodes"]

    def set_walltime(self, walltime: str) -> None:
        """Set the walltime

        This sets ``-W``.

        :param walltime: Time in hh:mm format, e.g. "10:00" for 10 hours,
                         if time is supplied in hh:mm:ss format, seconds
                         will be ignored and walltime will be set as ``hh:mm``
        """
        # For compatibility with other launchers, as explained in docstring
        if walltime:
            if len(walltime.split(":")) > 2:
                walltime = ":".join(walltime.split(":")[:2])
        self.walltime = walltime

    def set_smts(self, smts: int) -> None:
        """Set SMTs

        This sets ``-alloc_flags``. If the user sets
        SMT explicitly through ``-alloc_flags``, then that
        takes precedence.

        :param smts: SMT (e.g on Summit: 1, 2, or 4)
        """
        self.smts = smts

    def set_project(self, project: str) -> None:
        """Set the project

        This sets ``-P``.

        :param time: project name
        """
        if project:
            self.project = project

    def set_account(self, account: str) -> None:
        """Set the project

        this function is an alias for `set_project`.

        :param account: project name
        """
        self.set_project(account)

    def set_nodes(self, num_nodes: int) -> None:
        """Set the number of nodes for this batch job

        This sets ``-nnodes``.

        :param nodes: number of nodes
        """
        if num_nodes:
            self.batch_args["nnodes"] = str(int(num_nodes))

    def set_expert_mode_req(self, res_req: str, slots: int) -> None:
        """Set allocation for expert mode. This
        will activate expert mode (``-csm``) and
        disregard all other allocation options.

        This sets ``-csm -n slots -R res_req``

        :param res_req: specific resource requirements
        :param slots: number of resources to allocate
        """
        self.expert_mode = True
        self.batch_args["csm"] = "y"
        self.batch_args["R"] = res_req
        self.batch_args["n"] = str(slots)

    def set_hostlist(self, host_list: t.Union[str, t.List[str]]) -> None:
        """Specify the hostlist for this job

        :param host_list: hosts to launch on
        :raises TypeError: if not str or list of str
        """
        if isinstance(host_list, str):
            host_list = [host_list.strip()]
        if not isinstance(host_list, list):
            raise TypeError("host_list argument must be a list of strings")
        if not all(isinstance(host, str) for host in host_list):
            raise TypeError("host_list argument must be list of strings")
        self.batch_args["m"] = '"' + " ".join(host_list) + '"'

    def set_tasks(self, tasks: int) -> None:
        """Set the number of tasks for this job

        This sets ``-n``

        :param tasks: number of tasks
        """
        self.batch_args["n"] = str(int(tasks))

    def set_queue(self, queue: str) -> None:
        """Set the queue for this job

        :param queue: The queue to submit the job on
        """
        if queue:
            self.batch_args["q"] = queue

    def _format_alloc_flags(self) -> None:
        """Format ``alloc_flags`` checking if user already
        set it. Currently only adds SMT flag if missing
        and ``self.smts`` is set.
        """

        if self.smts:
            if "alloc_flags" not in self.batch_args.keys():
                self.batch_args["alloc_flags"] = f"smt{self.smts}"
            else:
                # Check if smt is in the flag, otherwise add it
                flags: t.List[str] = []
                if flags_arg := self.batch_args.get("alloc_flags", ""):
                    flags = flags_arg.strip('"').split()
                if not any(flag.startswith("smt") for flag in flags):
                    flags.append(f"smt{self.smts}")
                    self.batch_args["alloc_flags"] = " ".join(flags)

        # Check if alloc_flags has to be enclosed in quotes
        if "alloc_flags" in self.batch_args.keys():
            flags = []
            if flags_arg := self.batch_args.get("alloc_flags", ""):
                flags = flags_arg.strip('"').split()
            if len(flags) > 1:
                self.batch_args["alloc_flags"] = '"' + " ".join(flags) + '"'

    def format_batch_args(self) -> t.List[str]:
        """Get the formatted batch arguments for a preview

        :return: list of batch arguments for Qsub
        """
        opts = []

        self._format_alloc_flags()

        for opt, value in self.batch_args.items():
            if self.expert_mode and opt in self.easy_settings:
                continue

            prefix = "-"  # LSF only uses single dashses

            if not value:
                opts += [prefix + opt]
            else:
                opts += [" ".join((prefix + opt, str(value)))]

        return opts



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import typing as t

from pydantic import BaseModel, Field, PositiveInt

import smartsim._core.schemas.utils as _utils

# Black and Pylint disagree about where to put the `...`
# pylint: disable=multiple-statements

request_registry = _utils.SchemaRegistry["DragonRequest"]()


class DragonRequest(BaseModel): ...


class DragonRunRequestView(DragonRequest):
    exe: t.Annotated[str, Field(min_length=1)]
    exe_args: t.List[t.Annotated[str, Field(min_length=1)]] = []
    path: t.Annotated[str, Field(min_length=1)]
    nodes: PositiveInt = 1
    tasks: PositiveInt = 1
    tasks_per_node: PositiveInt = 1
    hostlist: t.Optional[t.Annotated[str, Field(min_length=1)]] = None
    output_file: t.Optional[t.Annotated[str, Field(min_length=1)]] = None
    error_file: t.Optional[t.Annotated[str, Field(min_length=1)]] = None
    env: t.Dict[str, t.Optional[str]] = {}
    name: t.Optional[t.Annotated[str, Field(min_length=1)]] = None
    pmi_enabled: bool = True


@request_registry.register("run")
class DragonRunRequest(DragonRunRequestView):
    current_env: t.Dict[str, t.Optional[str]] = {}

    def __str__(self) -> str:
        return str(DragonRunRequestView.parse_obj(self.dict(exclude={"current_env"})))


@request_registry.register("update_status")
class DragonUpdateStatusRequest(DragonRequest):
    step_ids: t.List[t.Annotated[str, Field(min_length=1)]]


@request_registry.register("stop")
class DragonStopRequest(DragonRequest):
    step_id: t.Annotated[str, Field(min_length=1)]


@request_registry.register("handshake")
class DragonHandshakeRequest(DragonRequest): ...


@request_registry.register("bootstrap")
class DragonBootstrapRequest(DragonRequest):
    address: t.Annotated[str, Field(min_length=1)]


@request_registry.register("shutdown")
class DragonShutdownRequest(DragonRequest):
    immediate: bool = True
    """Whether the server should shut down immediately, setting this to False means
    that the server will shut down when all jobs are terminated."""
    frontend_shutdown: bool = True
    """Whether the frontend will have to shut down or wait for external termination"""



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import importlib.util
import os
import shutil
import subprocess as sp
import sys
import typing as t
from argparse import ArgumentParser, Namespace
from pathlib import Path

from smartsim._core._install.buildenv import SetupError
from smartsim._core._install.builder import BuildError
from smartsim._core.utils import colorize
from smartsim.log import get_logger

SMART_LOGGER_FORMAT = "[%(name)s] %(levelname)s %(message)s"
logger = get_logger("Smart", fmt=SMART_LOGGER_FORMAT)


def get_install_path() -> Path:
    module_spec = importlib.util.find_spec("smartsim")

    if module_spec is None:
        raise SetupError("Could not import SmartSim") from None

    # find the path to the setup script
    package_dir = ""
    if module_spec.submodule_search_locations:
        package_dir = module_spec.submodule_search_locations[0]
    package_path = Path(package_dir).resolve()
    if not package_path.is_dir():
        raise SetupError("Could not find SmartSim installation site")

    return package_path


def color_bool(trigger: bool = True) -> str:
    _color = "green" if trigger else "red"
    return colorize(str(trigger), color=_color)


def pip(*args: str) -> None:
    cmd = (sys.executable, "-m", "pip") + args
    with sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE) as proc:
        _, err = proc.communicate()
        if int(proc.returncode) != 0:
            raise BuildError(
                f"`pip` returned with a non-zero exit code:\n{err.decode('utf-8')}"
            )


def clean(core_path: Path, _all: bool = False) -> int:
    """Remove pre existing installations of ML runtimes

    :param _all: Remove all non-python dependencies
    """

    build_temp = core_path / ".third-party"
    if build_temp.is_dir():
        shutil.rmtree(build_temp, ignore_errors=True)

    dragon_temp = core_path / ".dragon"
    if dragon_temp.is_dir():
        shutil.rmtree(dragon_temp, ignore_errors=True)
        logger.info("Successfully removed dragon installation")

    lib_path = core_path / "lib"
    if lib_path.is_dir():
        # remove RedisAI
        rai_path = lib_path / "redisai.so"
        if rai_path.is_file():
            rai_path.unlink()
            logger.info("Successfully removed existing RedisAI installation")

        backend_path = lib_path / "backends"
        if backend_path.is_dir():
            shutil.rmtree(backend_path, ignore_errors=True)
            logger.info("Successfully removed ML runtimes")

    bin_path = core_path / "bin"
    if bin_path.is_dir() and _all:
        files_to_remove = ["redis-server", "redis-cli", "keydb-server", "keydb-cli"]
        removed = False
        for _file in files_to_remove:
            file_path = bin_path.joinpath(_file)

            if file_path.is_file():
                removed = True
                file_path.unlink()
        if removed:
            logger.info("Successfully removed SmartSim database installation")

    return os.EX_OK


def get_db_path() -> t.Optional[Path]:
    bin_path = get_install_path() / "_core" / "bin"
    for option in bin_path.iterdir():
        if option.name in ("redis-cli", "keydb-cli"):
            return option
    return None


_CliHandler = t.Callable[[Namespace, t.List[str]], int]
_CliParseConfigurator = t.Callable[[ArgumentParser], None]


class MenuItemConfig:
    def __init__(
        self,
        cmd: str,
        description: str,
        handler: _CliHandler,
        configurator: t.Optional[_CliParseConfigurator] = None,
        is_plugin: bool = False,
    ):
        self.command = cmd
        self.description = description
        self.handler = handler
        self.configurator = configurator
        self.is_plugin = is_plugin



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import pathlib
import typing as t

from ..._core.launcher.step import Step
from ...entity import EntityList, Model
from ...error import SmartSimError
from ..launcher.launcher import Launcher

if t.TYPE_CHECKING:
    from ..utils.serialize import TStepLaunchMetaData


class _AnonymousBatchJob(EntityList[Model]):
    @staticmethod
    def _validate(model: Model) -> None:
        if model.batch_settings is None:
            msg = "Unable to create _AnonymousBatchJob without batch_settings"
            raise SmartSimError(msg)

    def __init__(self, model: Model) -> None:
        self._validate(model)
        super().__init__(model.name, model.path)
        self.entities = [model]
        self.batch_settings = model.batch_settings

    def _initialize_entities(self, **kwargs: t.Any) -> None: ...


def _look_up_launched_data(
    launcher: Launcher,
) -> t.Callable[[t.Tuple[str, Step]], "TStepLaunchMetaData"]:
    def _unpack_launched_data(data: t.Tuple[str, Step]) -> "TStepLaunchMetaData":
        # NOTE: we cannot assume that the name of the launched step
        # ``launched_step_name`` is equal to the name of the step referring to
        # the entity ``step.name`` as is the case when an entity list is
        # launched as a batch job
        launched_step_name, step = data
        launched_step_map = launcher.step_mapping[launched_step_name]
        out_file, err_file = step.get_output_files()
        return (
            launched_step_map.step_id,
            launched_step_map.task_id,
            launched_step_map.managed,
            out_file,
            err_file,
            pathlib.Path(step.meta.get("status_dir", step.cwd)),
        )

    return _unpack_launched_data



# BSD 2-Clause License
#
# Copyright (c) 2021-2024 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import abc
import logging
import pathlib
import typing as t

logger = logging.getLogger("TelemetryMonitor")


class Sink(abc.ABC):
    """Base class for output sinks. Represents a durable, read-only
    storage mechanism"""

    @abc.abstractmethod
    async def save(self, *args: t.Any) -> None:
        """Save the args passed to this method to the underlying sink

        :param args: variadic list of values to save
        """


class FileSink(Sink):
    """Telemetry sink that writes to a file"""

    def __init__(self, path: str) -> None:
        """Initialize the FileSink

        :param filename: path to a file backing this `Sink`
        """
        super().__init__()
        self._check_init(path)
        self._path = pathlib.Path(path)

    @staticmethod
    def _check_init(filename: str) -> None:
        """Validate initialization arguments and raise a ValueError
        if an invalid filename is passed

        :param filename: path to a file backing this `Sink`
        """
        if not filename:
            raise ValueError("No filename provided to FileSink")

    @property
    def path(self) -> pathlib.Path:
        """The path to the file this FileSink writes

        :return: path to a file backing this `Sink`
        """
        return self._path

    async def save(self, *args: t.Any) -> None:
        self._path.parent.mkdir(parents=True, exist_ok=True)

        with open(self._path, "a+", encoding="utf-8") as sink_fp:
            values = ",".join(map(str, args)) + "\n"
            sink_fp.write(values)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import signal
import socket
import sys
import tempfile
import typing as t
from pathlib import Path
from subprocess import STDOUT
from types import FrameType

import filelock
import psutil
from smartredis import Client, ConfigOptions
from smartredis.error import RedisConnectionError, RedisReplyError

from smartsim._core.utils.network import current_ip
from smartsim.error import SSInternalError
from smartsim.log import get_logger

logger = get_logger(__name__)

DBPID = None

# kill is not catchable
SIGNALS = [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGABRT]


def handle_signal(signo: int, _frame: t.Optional[FrameType]) -> None:
    if not signo:
        logger.warning("Received signal with no signo")
    cleanup()


def launch_db_model(client: Client, db_model: t.List[str]) -> str:
    """Parse options to launch model on local cluster

    :param client: SmartRedis client connected to local DB
    :param db_model: List of arguments defining the model
    :return: Name of model
    """
    parser = argparse.ArgumentParser("Set ML model on DB")
    parser.add_argument("--name", type=str)
    parser.add_argument("--file", type=str)
    parser.add_argument("--backend", type=str)
    parser.add_argument("--device", type=str)
    parser.add_argument("--devices_per_node", type=int, default=1)
    parser.add_argument("--first_device", type=int, default=0)
    parser.add_argument("--batch_size", type=int, default=0)
    parser.add_argument("--min_batch_size", type=int, default=0)
    parser.add_argument("--min_batch_timeout", type=int, default=0)
    parser.add_argument("--tag", type=str, default="")
    parser.add_argument("--inputs", nargs="+", default=None)
    parser.add_argument("--outputs", nargs="+", default=None)
    args = parser.parse_args(db_model)

    inputs = None
    outputs = None

    if args.inputs:
        inputs = list(args.inputs)
    if args.outputs:
        outputs = list(args.outputs)

    name = str(args.name)

    # devices_per_node being greater than one only applies to GPU devices
    if args.devices_per_node > 1 and args.device.lower() == "gpu":
        client.set_model_from_file_multigpu(
            name=name,
            model_file=args.file,
            backend=args.backend,
            first_gpu=args.first_device,
            num_gpus=args.devices_per_node,
            batch_size=args.batch_size,
            min_batch_size=args.min_batch_size,
            min_batch_timeout=args.min_batch_timeout,
            tag=args.tag,
            inputs=inputs,
            outputs=outputs,
        )
    else:
        client.set_model_from_file(
            name=name,
            model_file=args.file,
            backend=args.backend,
            device=args.device,
            batch_size=args.batch_size,
            min_batch_size=args.min_batch_size,
            min_batch_timeout=args.min_batch_timeout,
            tag=args.tag,
            inputs=inputs,
            outputs=outputs,
        )

    return name


def launch_db_script(client: Client, db_script: t.List[str]) -> str:
    """Parse options to launch script on local cluster

    :param client: SmartRedis client connected to local DB
    :param db_model: List of arguments defining the script
    :return: Name of model
    """
    parser = argparse.ArgumentParser("Set script on DB")
    parser.add_argument("--name", type=str)
    parser.add_argument("--func", type=str)
    parser.add_argument("--file", type=str)
    parser.add_argument("--backend", type=str)
    parser.add_argument("--device", type=str)
    parser.add_argument("--devices_per_node", type=int, default=1)
    parser.add_argument("--first_device", type=int, default=0)
    args = parser.parse_args(db_script)

    if args.file and args.func:
        raise ValueError("Both file and func cannot be provided.")

    if args.func:
        func = args.func.replace("\\n", "\n")
        if args.devices_per_node > 1 and args.device.lower() == "gpu":
            client.set_script_multigpu(
                args.name, func, args.first_device, args.devices_per_node
            )
        else:
            client.set_script(args.name, func, args.device)
    elif args.file:
        if args.devices_per_node > 1 and args.device.lower() == "gpu":
            client.set_script_from_file_multigpu(
                args.name, args.file, args.first_device, args.devices_per_node
            )
        else:
            client.set_script_from_file(args.name, args.file, args.device)
    else:
        raise ValueError("No file or func provided.")

    return str(args.name)


def main(
    network_interface: str,
    db_cpus: int,
    command: t.List[str],
    db_models: t.List[t.List[str]],
    db_scripts: t.List[t.List[str]],
    db_identifier: str,
) -> None:
    # pylint: disable=too-many-statements
    global DBPID  # pylint: disable=global-statement

    lo_address = current_ip("lo")
    ip_addresses = []
    if network_interface:
        try:
            ip_addresses = [
                current_ip(interface) for interface in network_interface.split(",")
            ]
        except ValueError as e:
            logger.warning(e)

    if all(lo_address == ip_address for ip_address in ip_addresses) or not ip_addresses:
        cmd = command + [f"--bind {lo_address}"]
    else:
        # bind to both addresses if the user specified a network
        # address that exists and is not the loopback address
        cmd = command + [f"--bind {lo_address} {' '.join(ip_addresses)}"]
        # pin source address to avoid random selection by Redis
        cmd += [f"--bind-source-addr {lo_address}"]

    # we generally want to catch all exceptions here as
    # if this process dies, the application will most likely fail
    try:
        hostname = socket.gethostname()
        filename = (
            f"colo_orc_{hostname}.log"
            if os.getenv("SMARTSIM_LOG_LEVEL") == "debug"
            else os.devnull
        )
        with open(filename, "w", encoding="utf-8") as file:
            process = psutil.Popen(cmd, stdout=file.fileno(), stderr=STDOUT)
            DBPID = process.pid
        # printing to stdout shell file for extraction
        print(f"__PID__{DBPID}__PID__", flush=True)

    except Exception as e:
        cleanup()
        logger.error(f"Failed to start database process: {str(e)}")
        raise SSInternalError("Colocated process failed to start") from e

    try:
        logger.debug(
            "\n\nColocated database information\n"
            f"\n\tIP Address(es): {' '.join(ip_addresses + [lo_address])}"
            f"\n\tCommand: {' '.join(cmd)}\n\n"
            f"\n\t# of Database CPUs: {db_cpus}"
            f"\n\tDatabase Identifier: {db_identifier}"
        )
    except Exception as e:
        cleanup()
        logger.error(f"Failed to start database process: {str(e)}")
        raise SSInternalError("Colocated process failed to start") from e

    def launch_models(client: Client, db_models: t.List[t.List[str]]) -> None:
        for i, db_model in enumerate(db_models):
            logger.debug("Uploading model")
            model_name = launch_db_model(client, db_model)
            logger.debug(f"Added model {model_name} ({i+1}/{len(db_models)})")

    def launch_db_scripts(client: Client, db_scripts: t.List[t.List[str]]) -> None:
        for i, db_script in enumerate(db_scripts):
            logger.debug("Uploading script")
            script_name = launch_db_script(client, db_script)
            logger.debug(f"Added script {script_name} ({i+1}/{len(db_scripts)})")

    try:
        if db_models or db_scripts:
            try:
                options = ConfigOptions.create_from_environment(db_identifier)
                client = Client(options, logger_name="SmartSim")
                launch_models(client, db_models)
                launch_db_scripts(client, db_scripts)
            except (RedisConnectionError, RedisReplyError) as ex:
                raise SSInternalError(
                    "Failed to set model or script, could not connect to database"
                ) from ex
            # Make sure we don't keep this around
            del client

    except Exception as e:
        cleanup()
        logger.error(f"Colocated database process failed: {str(e)}")
        raise SSInternalError("Colocated entrypoint raised an error") from e


def cleanup() -> None:
    try:
        logger.debug("Cleaning up colocated database")
        # attempt to stop the database process
        db_proc = psutil.Process(DBPID)
        db_proc.terminate()

    except psutil.NoSuchProcess:
        logger.warning("Couldn't find database process to kill.")

    except OSError as e:
        logger.warning(f"Failed to clean up colocated database gracefully: {str(e)}")
    finally:
        if LOCK.is_locked:
            LOCK.release()

        if os.path.exists(LOCK.lock_file):
            os.remove(LOCK.lock_file)


def register_signal_handlers() -> None:
    for sig in SIGNALS:
        signal.signal(sig, handle_signal)


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(
        prefix_chars="+", description="SmartSim Process Launcher"
    )
    arg_parser.add_argument(
        "+ifname", type=str, help="Network Interface name", default=""
    )
    arg_parser.add_argument(
        "+lockfile", type=str, help="Filename to create for single proc per host"
    )
    arg_parser.add_argument(
        "+db_cpus", type=int, default=2, help="Number of CPUs to use for DB"
    )

    arg_parser.add_argument(
        "+db_identifier", type=str, default="", help="Database Identifier"
    )

    arg_parser.add_argument("+command", nargs="+", help="Command to run")
    arg_parser.add_argument(
        "+db_model",
        nargs="+",
        action="append",
        default=[],
        help="Model to set on DB",
    )
    arg_parser.add_argument(
        "+db_script",
        nargs="+",
        action="append",
        default=[],
        help="Script to set on DB",
    )

    os.environ["PYTHONUNBUFFERED"] = "1"

    try:
        parsed_args = arg_parser.parse_args()
        tmp_lockfile = Path(tempfile.gettempdir()) / parsed_args.lockfile

        LOCK = filelock.FileLock(tmp_lockfile)
        LOCK.acquire(timeout=0.1)
        logger.debug(f"Starting colocated database on host: {socket.gethostname()}")

        # make sure to register the cleanup before we start
        # the proecss so our signaller will be able to stop
        # the database process.
        register_signal_handlers()

        main(
            parsed_args.ifname,
            parsed_args.db_cpus,
            parsed_args.command,
            parsed_args.db_model,
            parsed_args.db_script,
            parsed_args.db_identifier,
        )

    # gracefully exit the processes in the distributed application that
    # we do not want to have start a colocated process. Only one process
    # per node should be running.
    except filelock.Timeout:
        sys.exit(0)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import fileinput
import itertools
import json
import os
import os.path as osp
import time
import typing as t
from dataclasses import dataclass

from .._core.config import CONFIG
from ..error import SSDBFilesNotParseable
from ..log import get_logger
from ..settings.base import RunSettings
from .entity import SmartSimEntity

logger = get_logger(__name__)


class DBNode(SmartSimEntity):
    """DBNode objects are the entities that make up the orchestrator.
    Each database node can be launched in a cluster configuration
    and take launch multiple databases per node.

    To configure how each instance of the database operates, look
    into the smartsimdb.conf.
    """

    def __init__(
        self,
        name: str,
        path: str,
        run_settings: RunSettings,
        ports: t.List[int],
        output_files: t.List[str],
        db_identifier: str = "",
    ) -> None:
        """Initialize a database node within an orchestrator."""
        super().__init__(name, path, run_settings)
        self.ports = ports
        self._hosts: t.Optional[t.List[str]] = None

        if not output_files:
            raise ValueError("output_files cannot be empty")
        if not isinstance(output_files, list) or not all(
            isinstance(item, str) for item in output_files
        ):
            raise ValueError("output_files must be of type list[str]")
        self._output_files = output_files
        self.db_identifier = db_identifier

    @property
    def num_shards(self) -> int:
        if not hasattr(self.run_settings, "mpmd"):
            # return default number of shards if mpmd is not set
            return 1

        return len(self.run_settings.mpmd) + 1

    @property
    def host(self) -> str:
        try:
            (host,) = self.hosts
        except ValueError:
            raise ValueError(
                f"Multiple hosts detected for this DB Node: {', '.join(self.hosts)}"
            ) from None
        return host

    @property
    def hosts(self) -> t.List[str]:
        if not self._hosts:
            self._hosts = self._parse_db_hosts()
        return self._hosts

    def clear_hosts(self) -> None:
        self._hosts = None

    @property
    def is_mpmd(self) -> bool:
        if not hasattr(self.run_settings, "mpmd"):
            # missing mpmd property guarantees this is not an mpmd run
            return False

        return bool(self.run_settings.mpmd)

    def set_hosts(self, hosts: t.List[str]) -> None:
        self._hosts = [str(host) for host in hosts]

    def remove_stale_dbnode_files(self) -> None:
        """This function removes the .conf, .err, and .out files that
        have the same names used by this dbnode that may have been
        created from a previous experiment execution.
        """

        for port in self.ports:
            for conf_file in (
                osp.join(self.path, filename)
                for filename in self._get_cluster_conf_filenames(port)
            ):
                if osp.exists(conf_file):
                    os.remove(conf_file)

        for file_ending in [".err", ".out", ".mpmd"]:
            file_name = osp.join(self.path, self.name + file_ending)
            if osp.exists(file_name):
                os.remove(file_name)

        if self.is_mpmd:
            for file_ending in [".err", ".out"]:
                for shard_id in range(self.num_shards):
                    file_name = osp.join(
                        self.path, self.name + "_" + str(shard_id) + file_ending
                    )
                    if osp.exists(file_name):
                        os.remove(file_name)

    def _get_cluster_conf_filenames(self, port: int) -> t.List[str]:  # cov-lsf
        """Returns the .conf file name for the given port number

        This function should bu used if and only if ``_mpmd==True``

        :param port: port number
        :return: the dbnode configuration file name
        """
        if self.num_shards == 1:
            return [f"nodes-{self.name}-{port}.conf"]
        return [
            f"nodes-{self.name}_{shard_id}-{port}.conf"
            for shard_id in range(self.num_shards)
        ]

    @staticmethod
    def _parse_launched_shard_info_from_iterable(
        stream: t.Iterable[str], num_shards: t.Optional[int] = None
    ) -> "t.List[LaunchedShardData]":
        lines = (line.strip() for line in stream)
        lines = (line for line in lines if line)
        tokenized = (line.split(maxsplit=1) for line in lines)
        tokenized = (tokens for tokens in tokenized if len(tokens) > 1)
        shard_data_jsons = (
            kwjson for first, kwjson in tokenized if "SMARTSIM_ORC_SHARD_INFO" in first
        )
        shard_data_kwargs = (json.loads(kwjson) for kwjson in shard_data_jsons)
        shard_data: "t.Iterable[LaunchedShardData]" = (
            LaunchedShardData(**kwargs) for kwargs in shard_data_kwargs
        )
        if num_shards:
            shard_data = itertools.islice(shard_data, num_shards)
        return list(shard_data)

    @classmethod
    def _parse_launched_shard_info_from_files(
        cls, file_paths: t.List[str], num_shards: t.Optional[int] = None
    ) -> "t.List[LaunchedShardData]":
        with fileinput.FileInput(file_paths) as ifstream:
            return cls._parse_launched_shard_info_from_iterable(ifstream, num_shards)

    def get_launched_shard_info(self) -> "t.List[LaunchedShardData]":
        """Parse the launched database shard info from the output files

        :raises SSDBFilesNotParseable: if all shard info could not be found
        :return: The found launched shard info
        """
        ips: "t.List[LaunchedShardData]" = []
        trials = CONFIG.database_file_parse_trials
        interval = CONFIG.database_file_parse_interval
        output_files = [osp.join(self.path, file) for file in self._output_files]

        while len(ips) < self.num_shards and trials > 0:
            try:
                ips = self._parse_launched_shard_info_from_files(
                    output_files, self.num_shards
                )
            except FileNotFoundError:
                ...
            if len(ips) < self.num_shards:
                logger.debug("Waiting for output files to populate...")
                time.sleep(interval)
                trials -= 1

        if len(ips) < self.num_shards:
            msg = (
                f"Failed to parse the launched DB shard information from file(s) "
                f"{', '.join(output_files)}. Found the information for "
                f"{len(ips)} out of {self.num_shards} DB shards."
            )
            logger.error(msg)
            raise SSDBFilesNotParseable(msg)
        return ips

    def _parse_db_hosts(self) -> t.List[str]:
        """Parse the database hosts/IPs from the output files

        The IP address is preferred, but if hostname is only present
        then a lookup to /etc/hosts is done through the socket library.

        :raises SSDBFilesNotParseable: if host/ip could not be found
        :return: ip addresses | hostnames
        """
        return list({shard.hostname for shard in self.get_launched_shard_info()})


@dataclass(frozen=True)
class LaunchedShardData:
    """Data class to write and parse data about a launched database shard"""

    name: str
    hostname: str
    port: int
    cluster: bool

    @property
    def cluster_conf_file(self) -> t.Optional[str]:
        return f"nodes-{self.name}-{self.port}.conf" if self.cluster else None

    def to_dict(self) -> t.Dict[str, t.Any]:
        return dict(self.__dict__)



import os
import pathlib
import sys
import typing as t

from github import Github
from github.GitReleaseAsset import GitReleaseAsset

from smartsim._core._cli.utils import pip
from smartsim._core._install.builder import WebTGZ
from smartsim._core.config import CONFIG
from smartsim._core.utils.helpers import check_platform, is_crayex_platform
from smartsim.error.errors import SmartSimCLIActionCancelled
from smartsim.log import get_logger

logger = get_logger(__name__)


def create_dotenv(dragon_root_dir: pathlib.Path) -> None:
    """Create a .env file with required environment variables for the Dragon runtime"""
    dragon_root = str(dragon_root_dir)
    dragon_inc_dir = str(dragon_root_dir / "include")
    dragon_lib_dir = str(dragon_root_dir / "lib")
    dragon_bin_dir = str(dragon_root_dir / "bin")

    dragon_vars = {
        "DRAGON_BASE_DIR": dragon_root,
        "DRAGON_ROOT_DIR": dragon_root,  # note: same as base_dir
        "DRAGON_INCLUDE_DIR": dragon_inc_dir,
        "DRAGON_LIB_DIR": dragon_lib_dir,
        "DRAGON_VERSION": dragon_pin(),
        "PATH": dragon_bin_dir,
        "LD_LIBRARY_PATH": dragon_lib_dir,
    }

    lines = [f"{k}={v}\n" for k, v in dragon_vars.items()]

    if not CONFIG.dragon_dotenv.parent.exists():
        CONFIG.dragon_dotenv.parent.mkdir(parents=True)

    with CONFIG.dragon_dotenv.open("w", encoding="utf-8") as dotenv:
        dotenv.writelines(lines)


def python_version() -> str:
    """Return a formatted string used to filter release assets
    for the current python version"""
    return f"py{sys.version_info.major}.{sys.version_info.minor}"


def dragon_pin() -> str:
    """Return a string indicating the pinned major/minor version of the dragon
    package to install"""
    return "0.9"


def _platform_filter(asset_name: str) -> bool:
    """Return True if the asset name matches naming standard for current
    platform (Cray or non-Cray). Otherwise, returns False.

    :param asset_name: A value to inspect for keywords indicating a Cray EX asset
    :returns: True if supplied value is correct for current platform"""
    key = "crayex"
    is_cray = key in asset_name.lower()
    if is_crayex_platform():
        return is_cray
    return not is_cray


def _version_filter(asset_name: str) -> bool:
    """Return true if the supplied value contains a python version match

    :param asset_name: A value to inspect for keywords indicating a python version
    :returns: True if supplied value is correct for current python version"""
    return python_version() in asset_name


def _pin_filter(asset_name: str) -> bool:
    """Return true if the supplied value contains a dragon version pin match

    :param asset_name: A value to inspect for keywords indicating a dragon version
    :returns: True if supplied value is correct for current dragon version"""
    return f"dragon-{dragon_pin()}" in asset_name


def _get_release_assets() -> t.Collection[GitReleaseAsset]:
    """Retrieve a collection of available assets for all releases that satisfy
    the dragon version pin

    :returns: A collection of release assets"""
    git = Github()

    dragon_repo = git.get_repo("DragonHPC/dragon")

    if dragon_repo is None:
        raise SmartSimCLIActionCancelled("Unable to locate dragon repo")

    # find any releases matching our pinned version requirement
    tags = [tag for tag in dragon_repo.get_tags() if dragon_pin() in tag.name]
    # repo.get_latest_release fails if only pre-release results are returned
    pin_releases = list(dragon_repo.get_release(tag.name) for tag in tags)
    releases = sorted(pin_releases, key=lambda r: r.published_at, reverse=True)

    # take the most recent release for the given pin
    assets = releases[0].assets

    return assets


def filter_assets(assets: t.Collection[GitReleaseAsset]) -> t.Optional[GitReleaseAsset]:
    """Filter the available release assets so that HSTA agents are used
    when run on a Cray EX platform

    :param assets: The collection of dragon release assets to filter
    :returns: An asset meeting platform & version filtering requirements"""
    # Expect cray & non-cray assets that require a filter, e.g.
    # 'dragon-0.8-py3.9.4.1-bafaa887f.tar.gz',
    # 'dragon-0.8-py3.9.4.1-CRAYEX-ac132fe95.tar.gz'
    asset = next(
        (
            asset
            for asset in assets
            if _version_filter(asset.name)
            and _platform_filter(asset.name)
            and _pin_filter(asset.name)
        ),
        None,
    )
    return asset


def retrieve_asset_info() -> GitReleaseAsset:
    """Find a release asset that meets all necessary filtering criteria

    :param dragon_pin: identify the dragon version to install (e.g. dragon-0.8)
    :returns: A GitHub release asset"""
    assets = _get_release_assets()
    asset = filter_assets(assets)

    platform_result = check_platform()
    if not platform_result.is_cray:
        logger.warning("Installing Dragon without HSTA support")
        for msg in platform_result.failures:
            logger.warning(msg)

    if asset is None:
        raise SmartSimCLIActionCancelled("No dragon runtime asset available to install")

    logger.debug(f"Retrieved asset metadata: {asset}")
    return asset


def retrieve_asset(working_dir: pathlib.Path, asset: GitReleaseAsset) -> pathlib.Path:
    """Retrieve the physical file associated to a given GitHub release asset

    :param working_dir: location in file system where assets should be written
    :param asset: GitHub release asset to retrieve
    :returns: path to the downloaded asset"""
    if working_dir.exists() and list(working_dir.rglob("*.whl")):
        return working_dir

    archive = WebTGZ(asset.browser_download_url)
    archive.extract(working_dir)

    logger.debug(f"Retrieved {asset.browser_download_url} to {working_dir}")
    return working_dir


def install_package(asset_dir: pathlib.Path) -> int:
    """Install the package found in `asset_dir` into the current python environment

    :param asset_dir: path to a decompressed archive contents for a release asset"""
    wheels = asset_dir.rglob("*.whl")
    wheel_path = next(wheels, None)
    if not wheel_path:
        logger.error(f"No wheel found for package in {asset_dir}")
        return 1

    create_dotenv(wheel_path.parent)

    while wheel_path is not None:
        logger.info(f"Installing package: {wheel_path.absolute()}")

        try:
            pip("install", "--force-reinstall", str(wheel_path))
            wheel_path = next(wheels, None)
        except Exception:
            logger.error(f"Unable to install from {asset_dir}")
            return 1

    return 0


def cleanup(
    archive_path: t.Optional[pathlib.Path] = None,
) -> None:
    """Delete the downloaded asset and any files extracted during installation

    :param archive_path: path to a downloaded archive for a release asset"""
    if archive_path:
        archive_path.unlink(missing_ok=True)
        logger.debug(f"Deleted archive: {archive_path}")


def install_dragon(extraction_dir: t.Union[str, os.PathLike[str]]) -> int:
    """Retrieve a dragon runtime appropriate for the current platform
    and install to the current python environment
    :param extraction_dir: path for download and extraction of assets
    :returns: Integer return code, 0 for success, non-zero on failures"""
    if sys.platform == "darwin":
        logger.debug(f"Dragon not supported on platform: {sys.platform}")
        return 1

    extraction_dir = pathlib.Path(extraction_dir)
    filename: t.Optional[pathlib.Path] = None
    asset_dir: t.Optional[pathlib.Path] = None

    try:
        asset_info = retrieve_asset_info()
        asset_dir = retrieve_asset(extraction_dir, asset_info)

        return install_package(asset_dir)
    except Exception as ex:
        logger.error("Unable to install dragon runtime", exc_info=ex)
    finally:
        cleanup(filename)

    return 2


if __name__ == "__main__":
    sys.exit(install_dragon(CONFIG.core_path / ".dragon"))



#!/usr/bin/env python

# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import typing as t

from smartsim._core._cli.build import configure_parser as build_parser
from smartsim._core._cli.build import execute as build_execute
from smartsim._core._cli.clean import configure_parser as clean_parser
from smartsim._core._cli.clean import execute as clean_execute
from smartsim._core._cli.clean import execute_all as clobber_execute
from smartsim._core._cli.dbcli import execute as dbcli_execute
from smartsim._core._cli.info import execute as info_execute
from smartsim._core._cli.plugin import plugins
from smartsim._core._cli.site import execute as site_execute
from smartsim._core._cli.teardown import configure_parser as teardown_parser
from smartsim._core._cli.teardown import execute as teardown_execute
from smartsim._core._cli.utils import MenuItemConfig
from smartsim._core._cli.validate import configure_parser as validate_parser
from smartsim._core._cli.validate import execute as validate_execute


class SmartCli:
    def __init__(self, menu: t.List[MenuItemConfig]) -> None:
        self.menu: t.Dict[str, MenuItemConfig] = {}
        self.parser = argparse.ArgumentParser(
            prog="smart",
            description="SmartSim command line interface",
        )

        self.subparsers = self.parser.add_subparsers(
            dest="command",
            required=True,
            metavar="<command>",
            help="Available commands",
        )

        self.register_menu_items(menu)
        self.register_menu_items([plugin() for plugin in plugins])

    def execute(self, cli_args: t.List[str]) -> int:
        if len(cli_args) < 2:
            self.parser.print_help()
            return os.EX_USAGE

        app_args = cli_args[1:]  # exclude the path to executable
        subcommand = cli_args[1]  # first positional arg is the subcommand

        menu_item = self.menu.get(subcommand, None)
        if not menu_item:
            self.parser.print_help()
            return os.EX_USAGE

        args = argparse.Namespace()
        unparsed_args = []

        if menu_item.is_plugin:
            unparsed_args = app_args[1:]
        else:
            args = self.parser.parse_args(app_args)

        return menu_item.handler(args, unparsed_args)

    def _register_menu_item(self, item: MenuItemConfig) -> None:
        parser = self.subparsers.add_parser(
            item.command, description=item.description, help=item.description
        )
        if item.configurator:
            item.configurator(parser)

        if item.command in self.menu:
            raise ValueError(f"{item.command} cannot overwrite existing CLI command")

        self.menu[item.command] = item

    def register_menu_items(self, menu_items: t.List[MenuItemConfig]) -> None:
        for item in menu_items:
            self._register_menu_item(item)


def default_cli() -> SmartCli:
    menu = [
        MenuItemConfig(
            "build",
            "Build SmartSim dependencies (Redis, RedisAI, Dragon, ML runtimes)",
            build_execute,
            build_parser,
        ),
        MenuItemConfig(
            "clean",
            "Remove previous ML runtime installation",
            clean_execute,
            clean_parser,
        ),
        MenuItemConfig(
            "dbcli",
            "Print the path to the redis-cli binary",
            dbcli_execute,
        ),
        MenuItemConfig(
            "site",
            "Print the installation site of SmartSim",
            site_execute,
        ),
        MenuItemConfig(
            "clobber",
            "Remove all previous dependency installations",
            clobber_execute,
        ),
        MenuItemConfig(
            "validate",
            "Run a simple SmartSim experiment to confirm that it is built correctly",
            validate_execute,
            validate_parser,
        ),
        MenuItemConfig(
            "info",
            "Display information about the current SmartSim installation",
            info_execute,
        ),
        MenuItemConfig(
            "teardown",
            "Clean up allocated resources after an experiment terminates",
            teardown_execute,
            teardown_parser,
        ),
    ]

    return SmartCli(menu)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=invalid-name

import importlib.metadata
import os
import platform
import subprocess
import sys
import typing as t
from pathlib import Path
from typing import Iterable

# NOTE: This will be imported by setup.py and hence no
#       smartsim related items or non-standand library
#       items should be imported here.

# TODO: pkg_resources has been deprecated by PyPA. Currently we use it for its
#       packaging implementation, as we cannot assume a user will have `packaging`
#       prior to `pip install` time. We really only use pkg_resources for their
#       vendored version of `packaging.version.Version` so we should probably try
#       to remove
# https://setuptools.pypa.io/en/latest/pkg_resources.html

# isort: off
import pkg_resources
from pkg_resources import packaging  # type: ignore

# isort: on

Version = packaging.version.Version
InvalidVersion = packaging.version.InvalidVersion
DbEngine = t.Literal["REDIS", "KEYDB"]


class SetupError(Exception):
    """A simple exception class for errors in _install.buildenv file.
    This is primarily used to interrupt the setup.py build in case
    of a failure, and is caught frequently in the CLI which attempts
    to suppress and log the errors thrown here.

    One note is that this error must be imported from here to be
    caught and not defined anywhere else. Since we use this class
    in the installation, having a separate error file would mean
    another module to manually import into setup.py

    See setup.py for more
    """


class VersionConflictError(SetupError):
    """An error for when version numbers of some library/package/program/etc
    do not match and build may not be able to continue
    """

    def __init__(
        self,
        name: str,
        current_version: "Version_",
        target_version: "Version_",
        msg: t.Optional[str] = None,
    ) -> None:
        if msg is None:
            msg = (
                f"Incompatible version for {name} detected: "
                f"{name} {target_version} requested but {name} {current_version} "
                "installed."
            )
        super().__init__(msg)
        self.name = name
        self.current_version = current_version
        self.target_version = target_version


# so as to not conflict with pkg_resources.packaging.version.Version
# pylint: disable-next=invalid-name
class Version_(str):
    """A subclass of pkg_resources.packaging.version.Version that
    includes some helper methods for comparing versions.
    """

    @staticmethod
    def _convert_to_version(
        vers: t.Union[
            str, Iterable[packaging.version.Version], packaging.version.Version
        ],
    ) -> t.Any:
        if isinstance(vers, Version):
            return vers
        if isinstance(vers, str):
            return Version(vers)
        if isinstance(vers, Iterable):
            return Version(".".join((str(item) for item in vers)))

        raise InvalidVersion(vers)

    @property
    def major(self) -> int:
        # Version(self).major doesn't work for all Python distributions
        # see https://github.com/lebedov/python-pdfbox/issues/28
        return int(pkg_resources.parse_version(self).base_version.split(".")[0])

    @property
    def minor(self) -> int:
        return int(pkg_resources.parse_version(self).base_version.split(".")[1])

    @property
    def micro(self) -> int:
        return int(pkg_resources.parse_version(self).base_version.split(".")[2])

    @property
    def patch(self) -> str:
        # return micro with string modifier i.e. 1.2.3+cpu -> 3+cpu
        return str(pkg_resources.parse_version(self)).split(".")[2]

    def __gt__(self, cmp: t.Any) -> bool:
        try:
            return bool(Version(self).__gt__(self._convert_to_version(cmp)))
        except InvalidVersion:
            return super().__gt__(cmp)

    def __lt__(self, cmp: t.Any) -> bool:
        try:
            return bool(Version(self).__lt__(self._convert_to_version(cmp)))
        except InvalidVersion:
            return super().__lt__(cmp)

    def __eq__(self, cmp: t.Any) -> bool:
        try:
            return bool(Version(self).__eq__(self._convert_to_version(cmp)))
        except InvalidVersion:
            return super().__eq__(cmp)

    def __ge__(self, cmp: t.Any) -> bool:
        try:
            return bool(Version(self).__ge__(self._convert_to_version(cmp)))
        except InvalidVersion:
            return super().__ge__(cmp)

    def __le__(self, cmp: t.Any) -> bool:
        try:
            return bool(Version(self).__le__(self._convert_to_version(cmp)))
        except InvalidVersion:
            return super().__le__(cmp)

    def __hash__(self) -> int:
        return hash(Version(self))


def get_env(var: str, default: str) -> str:
    return os.environ.get(var, default)


class RedisAIVersion(Version_):
    """A subclass of Version_ that holds the dependency sets for RedisAI

    this class serves two purposes:

    1. It is used to populate the [ml] ``extras_require`` of the setup.py.
    This is because the RedisAI version will determine which ML based
    dependencies are required.

    2. Used to set the default values for PyTorch, TF, and ONNX
    given the SMARTSIM_REDISAI env var set by the user.

    NOTE: Torch requires additional information depending on whether
    CPU or GPU support is requested
    """

    defaults = {
        "1.2.7": {
            "tensorflow": "2.13.1",
            "onnx": "1.14.1",
            "skl2onnx": "1.16.0",
            "onnxmltools": "1.12.0",
            "scikit-learn": "1.3.2",
            "torch": "2.0.1",
            "torch_cpu_suffix": "+cpu",
            "torch_cuda_suffix": "+cu117",
            "torchvision": "0.15.2",
        },
    }

    def __init__(self, vers: str) -> None:  # pylint: disable=super-init-not-called
        min_rai_version = min(Version_(ver) for ver in self.defaults)
        if min_rai_version > vers:
            raise SetupError(
                f"RedisAI version must be greater than or equal to {min_rai_version}"
            )
        if vers not in self.defaults:
            if vers.startswith("1.2"):
                # resolve to latest version for 1.2.x
                # the str representation will still be 1.2.x
                self.version = "1.2.7"
            else:
                raise SetupError(
                    (
                        f"Invalid RedisAI version {vers}. Options are "
                        f"{self.defaults.keys()}"
                    )
                )
        else:
            self.version = vers

    def __getattr__(self, name: str) -> str:
        try:
            return self.defaults[self.version][name]
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'\n\n"
                "This is likely a problem with the SmartSim build process;"
                "if this problem persists please log a new issue at "
                "https://github.com/CrayLabs/SmartSim/issues "
                "or get in contact with us at "
                "https://www.craylabs.org/docs/community.html"
            ) from None

    def get_defaults(self) -> t.Dict[str, str]:
        return self.defaults[self.version].copy()


class Versioner:
    """Versioner is responsible for managing all the versions
    within SmartSim including SmartSim itself.

    SmartSim's version is written into version.py upon pip install
    by using this class in setup.py. By setting SMARTSIM_SUFFIX,
    the version will be written as a "dirty" version with the
    git-sha appended.
    i.e.
        export SMARTSIM_SUFFIX=nightly
        pip install -e .
        smartsim.__version__ == 0.3.2+nightly-3fe23ff

    Versioner manages third party dependencies and their versions
    as well. These versions are used by the ``smart`` cli in the
    ``smart build`` command to determine which dependency versions
    to look for and download.

    Default versions for SmartSim, SmartRedis, Redis, and RedisAI are
    all set here. Setting a default version for RedisAI also dictates
    default versions of the machine learning libraries.
    """

    # compatible Python version
    PYTHON_MIN = Version_("3.9.0")

    # Versions
    SMARTSIM = Version_(get_env("SMARTSIM_VERSION", "0.7.0"))
    SMARTREDIS = Version_(get_env("SMARTREDIS_VERSION", "0.5.3"))
    SMARTSIM_SUFFIX = get_env("SMARTSIM_SUFFIX", "")

    # Redis
    REDIS = Version_(get_env("SMARTSIM_REDIS", "7.2.4"))
    REDIS_URL = get_env("SMARTSIM_REDIS_URL", "https://github.com/redis/redis.git/")
    REDIS_BRANCH = get_env("SMARTSIM_REDIS_BRANCH", REDIS)

    # RedisAI
    REDISAI = RedisAIVersion(get_env("SMARTSIM_REDISAI", "1.2.7"))
    REDISAI_URL = get_env(
        "SMARTSIM_REDISAI_URL", "https://github.com/RedisAI/RedisAI.git/"
    )
    REDISAI_BRANCH = get_env("SMARTSIM_REDISAI_BRANCH", f"v{REDISAI}")

    # ML/DL (based on RedisAI version defaults)
    # torch can be set by the user because we download that for them
    TORCH = Version_(get_env("SMARTSIM_TORCH", REDISAI.torch))
    TORCHVISION = Version_(get_env("SMARTSIM_TORCHVIS", REDISAI.torchvision))
    TORCH_CPU_SUFFIX = Version_(get_env("TORCH_CPU_SUFFIX", REDISAI.torch_cpu_suffix))
    TORCH_CUDA_SUFFIX = Version_(
        get_env("TORCH_CUDA_SUFFIX", REDISAI.torch_cuda_suffix)
    )

    # TensorFlow and ONNX only use the defaults, but these are not built into
    # the RedisAI package and therefore the user is free to pick other versions.
    TENSORFLOW = Version_(REDISAI.tensorflow)
    ONNX = Version_(REDISAI.onnx)

    def as_dict(self, db_name: DbEngine = "REDIS") -> t.Dict[str, t.Tuple[str, ...]]:
        pkg_map = {
            "SMARTSIM": self.SMARTSIM,
            "SMARTREDIS": self.SMARTREDIS,
            db_name: self.REDIS,
            "REDISAI": self.REDISAI,
            "TORCH": self.TORCH,
            "TENSORFLOW": self.TENSORFLOW,
            "ONNX": self.ONNX,
        }
        return {"Packages": tuple(pkg_map), "Versions": tuple(pkg_map.values())}

    def ml_extras_required(self) -> t.Dict[str, t.List[str]]:
        """Optional ML/DL dependencies we suggest for the user.

        The defaults are based on the RedisAI version
        """
        ml_defaults = self.REDISAI.get_defaults()

        # remove torch-related fields as they are subject to change
        # by having the user change hardware (cpu/gpu)
        _torch_fields = [
            "torch",
            "torchvision",
            "torch_cpu_suffix",
            "torch_cuda_suffix",
        ]
        for field in _torch_fields:
            ml_defaults.pop(field)

        return {"ml": [f"{lib}=={vers}" for lib, vers in ml_defaults.items()]}

    @staticmethod
    def get_sha(setup_py_dir: Path) -> str:
        """Get the git sha of the current branch"""
        try:
            rev_cmd = ["git", "rev-parse", "HEAD"]
            git_rev = subprocess.check_output(rev_cmd, cwd=setup_py_dir.absolute())
            sha = git_rev.decode("ascii").strip()

            return sha[:7]
        except Exception:
            # return empty string if not in git-repo
            return ""

    def write_version(self, setup_py_dir: Path) -> str:
        """
        Write version info to version.py

        Use git_sha in the case where smartsim suffix is set in the environment
        """
        version = str(self.SMARTSIM)

        if self.SMARTSIM_SUFFIX:
            version += f"+{self.SMARTSIM_SUFFIX}"

            # wheel build (python -m build) won't include git sha
            if git_sha := self.get_sha(setup_py_dir):
                version += f".{git_sha}"

        version_file_path = setup_py_dir / "smartsim" / "version.py"
        with open(version_file_path, "w", encoding="utf-8") as version_file:
            version_file.write("# This file is automatically generated by setup.py\n")
            version_file.write("# do not edit this file manually.\n\n")

            version_file.write(f"__version__ = '{version}'\n")
        return version


class BuildEnv:
    """Environment for building third-party dependencies

    BuildEnv provides a method for configuring how the third-party
    dependencies within SmartSim are built, namely Redis/KeyDB
    and RedisAI.

    The environment variables listed here can be set to control the
    Redis build in the pip wheel build as well as the Redis and RedisAI
    build executed by the CLI.

    Build tools are also checked for here and if they are not found
    then a SetupError is raised.

    Adding the -v flag to ``smart build`` will show the build environment
    being used.
    """

    # Compiler overrides
    CC = os.environ.get("CC", "gcc")
    CXX = os.environ.get("CXX", "g++")
    CFLAGS = os.environ.get("CFLAGS", "")
    CXXFLAGS = os.environ.get("CXXFLAGS", "")

    # build overrides
    MALLOC = os.environ.get("MALLOC", "libc")
    JOBS = int(os.environ.get("BUILD_JOBS", 1))

    # check for CC/GCC/ETC
    CHECKS = int(os.environ.get("NO_CHECKS", 0))
    PLATFORM = sys.platform

    def __init__(self, checks: bool = True) -> None:
        if checks:
            self.check_dependencies()

    def check_dependencies(self) -> None:
        deps = ["git", "git-lfs", "make", "wget", "cmake", self.CC, self.CXX]
        if int(self.CHECKS) == 0:
            for dep in deps:
                self.check_build_dependency(dep)

    def __call__(self) -> t.Dict[str, str]:
        # return the build env for the build process
        env = os.environ.copy()
        env.update(
            {
                "CC": self.CC,
                "CXX": self.CXX,
                "CFLAGS": self.CFLAGS,
                "CXXFLAGS": self.CXXFLAGS,
            }
        )
        return env

    def as_dict(self) -> t.Dict[str, t.List[str]]:
        variables: t.List[str] = [
            "CC",
            "CXX",
            "CFLAGS",
            "CXXFLAGS",
            "MALLOC",
            "JOBS",
            "PYTHON_VERSION",
            "PLATFORM",
        ]
        values: t.List[str] = [
            self.CC,
            self.CXX,
            self.CFLAGS,
            self.CXXFLAGS,
            self.MALLOC,
            str(self.JOBS),
            self.python_version,
            self.PLATFORM,
        ]
        env = {"Environment": variables, "Values": values}
        return env

    @property
    def python_version(self) -> str:
        return platform.python_version()

    @staticmethod
    def is_compatible_python(python_min: float) -> bool:
        """Detect if system Python is too old"""
        sys_py = sys.version_info
        system_python = Version_(f"{sys_py.major}.{sys_py.minor}.{sys_py.micro}")
        return system_python > python_min

    @classmethod
    def is_windows(cls) -> bool:
        return cls.PLATFORM in ["win32", "cygwin", "msys"]

    @classmethod
    def is_macos(cls) -> bool:
        return cls.PLATFORM == "darwin"

    @staticmethod
    def get_cudnn_env() -> t.Optional[t.Dict[str, str]]:
        """Collect the environment variables needed for Caffe (Pytorch)
        and throw an error if they are not found

        Specifically make sure to set at least one set of:
            - CUDNN_LIBRARY and CUDNN_INCLUDE_DIR
                or
            - CUDNN_LIBRARY_PATH and CUDNN_INCLUDE_PATH
        """
        env = {
            "CUDNN_LIBRARY": os.environ.get("CUDNN_LIBRARY", "env-var-not-found"),
            "CUDNN_INCLUDE_DIR": os.environ.get(
                "CUDNN_INCLUDE_DIR", "env-var-not-found"
            ),
            "CUDNN_LIBRARY_PATH": os.environ.get(
                "CUDNN_LIBRARY_PATH", "env-var-not-found"
            ),
            "CUDNN_INCLUDE_PATH": os.environ.get(
                "CUDNN_INCLUDE_PATH", "env-var-not-found"
            ),
        }
        torch_cudnn_vars = ["CUDNN_LIBRARY", "CUDNN_INCLUDE_DIR"]
        caffe_cudnn_vars = ["CUDNN_INCLUDE_PATH", "CUDNN_LIBRARY_PATH"]

        torch_set = all(var in os.environ for var in torch_cudnn_vars)
        caffe_set = all(var in os.environ for var in caffe_cudnn_vars)

        # check for both sets, if only one exists, set the other
        # this handles older versions which use different env vars
        if not torch_set and not caffe_set:
            return None  # return None as we don't want to raise a warning here
        if torch_set and not caffe_set:
            env["CUDNN_INCLUDE_PATH"] = env["CUDNN_INCLUDE_DIR"]
            env["CUDNN_LIBRARY_PATH"] = env["CUDNN_LIBRARY"]
        elif caffe_set and not torch_set:
            env["CUDNN_INCLUDE_DIR"] = env["CUDNN_INCLUDE_PATH"]
            env["CUDNN_LIBRARY"] = env["CUDNN_LIBRARY_PATH"]
        return env

    @staticmethod
    def check_build_dependency(command: str) -> None:
        # TODO expand this to parse and check versions.
        try:
            subprocess.check_call(
                [command, "--version"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        except OSError:
            raise SetupError(f"{command} must be installed to build SmartSim") from None

    @classmethod
    def check_installed(
        cls, package: str, version: t.Optional[Version_] = None
    ) -> bool:
        """Check if a package is installed. If version is provided, check if
        it's a compatible version. (major and minor the same)
        """
        try:
            installed = cls.get_py_package_version(package)
        except importlib.metadata.PackageNotFoundError:
            return False
        if version:
            # detect if major or minor versions differ
            if installed.major != version.major or installed.minor != version.minor:
                raise VersionConflictError(package, installed, version)
        return True

    @staticmethod
    def get_py_package_version(package: str) -> Version_:
        return Version_(importlib.metadata.version(package))



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
A file of helper functions for SmartSim
"""
import base64
import collections.abc
import os
import signal
import subprocess
import typing as t
import uuid
from datetime import datetime
from functools import lru_cache
from pathlib import Path
from shutil import which

from smartsim._core._install.builder import TRedisAIBackendStr as _TRedisAIBackendStr

if t.TYPE_CHECKING:
    from types import FrameType


_TSignalHandlerFn = t.Callable[[int, t.Optional["FrameType"]], object]


def unpack_db_identifier(db_id: str, token: str) -> t.Tuple[str, str]:
    """Unpack the unformatted database identifier
    and format for env variable suffix using the token
    :param db_id: the unformatted database identifier eg. identifier_1
    :param token: character to use to construct the db suffix
    :return: db id suffix and formatted db_id e.g. ("_identifier_1", "identifier_1")
    """

    if db_id == "orchestrator":
        return "", ""
    db_name_suffix = token + db_id
    return db_name_suffix, db_id


def unpack_colo_db_identifier(db_id: str) -> str:
    """Create database identifier suffix for colocated database

    :param db_id: the unformatted database identifier
    :return: db suffix
    """
    return "_" + db_id if db_id else ""


def create_short_id_str() -> str:
    return str(uuid.uuid4())[:7]


def create_lockfile_name() -> str:
    """Generate a unique lock filename using UUID"""
    lock_suffix = create_short_id_str()
    return f"smartsim-{lock_suffix}.lock"


@lru_cache(maxsize=20, typed=False)
def check_dev_log_level() -> bool:
    lvl = os.environ.get("SMARTSIM_LOG_LEVEL", "")
    return lvl == "developer"


def fmt_dict(value: t.Dict[str, t.Any]) -> str:
    fmt_str = ""
    for k, v in value.items():
        fmt_str += "\t" + str(k) + " = " + str(v)
        fmt_str += "\n" if k != list(value.keys())[-1] else ""
    return fmt_str


def get_base_36_repr(positive_int: int) -> str:
    """Converts a positive integer to its base 36 representation

    :param positive_int: the positive integer to convert
    :return: base 36 representation of the given positive int
    """
    digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    result = []

    while positive_int:
        next_digit = digits[positive_int % 36]
        result.append(next_digit)
        positive_int //= 36

    return "".join(reversed(result))


def expand_exe_path(exe: str) -> str:
    """Takes an executable and returns the full path to that executable

    :param exe: executable or file
    :raises TypeError: if file is not an executable
    :raises FileNotFoundError: if executable cannot be found
    """

    # which returns none if not found
    in_path = which(exe)
    if not in_path:
        if os.path.isfile(exe) and os.access(exe, os.X_OK):
            return os.path.abspath(exe)
        if os.path.isfile(exe) and not os.access(exe, os.X_OK):
            raise TypeError(f"File, {exe}, is not an executable")
        raise FileNotFoundError(f"Could not locate executable {exe}")
    return os.path.abspath(in_path)


def is_valid_cmd(command: t.Union[str, None]) -> bool:
    try:
        if command:
            expand_exe_path(command)
            return True
    except (TypeError, FileNotFoundError):
        return False

    return False


color2num = {
    "gray": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
    "crimson": 38,
}


def colorize(
    string: str, color: str, bold: bool = False, highlight: bool = False
) -> str:
    """
    Colorize a string.
    This function was originally written by John Schulman.
    And then borrowed from spinningup
    https://github.com/openai/spinningup/blob/master/spinup/utils/logx.py
    """
    attr = []
    num = color2num[color]
    if highlight:
        num += 10
    attr.append(str(num))
    if bold:
        attr.append("1")
    return f"\x1b[{';'.join(attr)}m{string}\x1b[0m"


def delete_elements(dictionary: t.Dict[str, t.Any], key_list: t.List[str]) -> None:
    """Delete elements from a dictionary.
    :param dictionary: the dictionary from which the elements must be deleted.
    :param key_list: the list of keys to delete from the dictionary.
    """
    for key in key_list:
        if key in dictionary:
            del dictionary[key]


def cat_arg_and_value(arg_name: str, value: str) -> str:
    """Concatenate a command line argument and its value

    This function returns ``arg_name`` and ``value
    concatenated in the best possible way for a command
    line execution, namely:
    - if arg_name starts with `--` (e.g. `--arg`):
      `arg_name=value` is returned (i.e. `--arg=val`)
    - if arg_name starts with `-` (e.g. `-a`):
      `arg_name value` is returned (i.e. `-a val`)
    - if arg_name does not start with `-` and it is a
      long option (e.g. `arg`):
      `--arg_name=value` (i.e., `--arg=val`)
    - if arg_name does not start with `-` and it is a
      short option (e.g. `a`):
      `-arg_name=value` (i.e., `-a val`)

    :param arg_name: the command line argument name
    :param value: the command line argument value
    """

    if arg_name.startswith("--"):
        return f"{arg_name}={value}"
    if arg_name.startswith("-"):
        return f"{arg_name} {value}"
    if len(arg_name) == 1:
        return f"-{arg_name} {value}"

    return f"--{arg_name}={value}"


def _installed(base_path: Path, backend: str) -> bool:
    """
    Check if a backend is available for the RedisAI module.
    """
    backend_key = f"redisai_{backend}"
    backend_path = base_path / backend_key / f"{backend_key}.so"
    backend_so = Path(os.environ.get("RAI_PATH", backend_path)).resolve()

    return backend_so.is_file()


def redis_install_base(backends_path: t.Optional[str] = None) -> Path:
    # pylint: disable-next=import-outside-toplevel
    from ..._core.config import CONFIG

    base_path = Path(backends_path) if backends_path else CONFIG.lib_path / "backends"
    return base_path


def installed_redisai_backends(
    backends_path: t.Optional[str] = None,
) -> t.Set[_TRedisAIBackendStr]:
    """Check which ML backends are available for the RedisAI module.

    The optional argument ``backends_path`` is needed if the backends
    have not been built as part of the SmartSim building process (i.e.
    they have not been built by invoking `smart build`). In that case
    ``backends_path`` should point to the directory containing e.g.
    the backend directories (`redisai_tensorflow`, `redisai_torch`,
    `redisai_onnxruntime`, or `redisai_tflite`).

    :param backends_path: path containing backends
    :return: list of installed RedisAI backends
    """
    # import here to avoid circular import
    base_path = redis_install_base(backends_path)
    backends: t.Set[_TRedisAIBackendStr] = {
        "tensorflow",
        "torch",
        "onnxruntime",
        "tflite",
    }

    return {backend for backend in backends if _installed(base_path, backend)}


def get_ts_ms() -> int:
    """Return the current timestamp (accurate to milliseconds) cast to an integer"""
    return int(datetime.now().timestamp() * 1000)


def encode_cmd(cmd: t.Sequence[str]) -> str:
    """Transform a standard command list into an encoded string safe for providing as an
    argument to a proxy entrypoint
    """
    if not cmd:
        raise ValueError("Invalid cmd supplied")

    ascii_cmd = "|".join(cmd).encode("ascii")
    encoded_cmd = base64.b64encode(ascii_cmd).decode("ascii")
    return encoded_cmd


def decode_cmd(encoded_cmd: str) -> t.List[str]:
    """Decode an encoded command string to the original command list format"""
    if not encoded_cmd.strip():
        raise ValueError("Invalid cmd supplied")

    decoded_cmd = base64.b64decode(encoded_cmd.encode("ascii"))
    cleaned_cmd = decoded_cmd.decode("ascii").split("|")

    return cleaned_cmd


def check_for_utility(util_name: str) -> str:
    """Check for existence of the provided CLI utility.

    :param util_name: CLI utility to locate
    :returns: Full path to executable if found. Otherwise, empty string"""
    utility = ""

    try:
        utility = expand_exe_path(util_name)
    except FileNotFoundError:
        ...

    return utility


def execute_platform_cmd(cmd: str) -> t.Tuple[str, int]:
    """Execute the platform check command as a subprocess

    :param cmd: the command to execute
    :returns: True if platform is cray ex, False otherwise"""
    process = subprocess.run(
        cmd.split(),
        capture_output=True,
        check=False,
    )
    return process.stdout.decode("utf-8"), process.returncode


class CrayExPlatformResult:
    locate_msg = "Unable to locate `{0}`."

    def __init__(self, ldconfig: t.Optional[str], fi_info: t.Optional[str]) -> None:
        self.ldconfig: t.Optional[str] = ldconfig
        self.fi_info: t.Optional[str] = fi_info
        self.has_pmi: bool = False
        self.has_pmi2: bool = False
        self.has_cxi: bool = False

    @property
    def has_ldconfig(self) -> bool:
        return bool(self.ldconfig)

    @property
    def has_fi_info(self) -> bool:
        return bool(self.fi_info)

    @property
    def is_cray(self) -> bool:
        return all(
            (
                self.has_ldconfig,
                self.has_fi_info,
                self.has_pmi,
                self.has_pmi2,
                self.has_cxi,
            )
        )

    @property
    def failures(self) -> t.List[str]:
        """Return a list of messages describing all failed validations"""
        failure_messages = []

        if not self.has_ldconfig:
            failure_messages.append(self.locate_msg.format("ldconfig"))

        if not self.has_fi_info:
            failure_messages.append(self.locate_msg.format("fi_info"))

        if self.has_ldconfig and self.has_fi_info:
            if not self.has_pmi:
                failure_messages.append(self.locate_msg.format("pmi.so"))
            if not self.has_pmi2:
                failure_messages.append(self.locate_msg.format("pmi2.so"))
            if not self.has_cxi:
                failure_messages.append(self.locate_msg.format("cxi.so"))

        return failure_messages


def check_platform() -> CrayExPlatformResult:
    """Returns True if the current platform is identified as Cray EX and
    HSTA-aware dragon package can be installed, False otherwise.

    :returns: True if current platform is Cray EX, False otherwise"""

    # ldconfig -p | grep cray | grep pmi.so &&
    # ldconfig -p | grep cray | grep pmi2.so &&
    # fi_info | grep cxi

    ldconfig = check_for_utility("ldconfig")
    fi_info = check_for_utility("fi_info")

    result = CrayExPlatformResult(ldconfig, fi_info)
    if not all((result.has_ldconfig, result.has_fi_info)):
        return result

    ldconfig1 = f"{ldconfig} -p"
    ldc_out1, _ = execute_platform_cmd(ldconfig1)
    candidates = [x for x in ldc_out1.split("\n") if "cray" in x]
    result.has_pmi = any(x for x in candidates if "pmi.so" in x)

    ldconfig2 = f"{ldconfig} -p"
    ldc_out2, _ = execute_platform_cmd(ldconfig2)
    candidates = [x for x in ldc_out2.split("\n") if "cray" in x]
    result.has_pmi2 = any(x for x in candidates if "pmi2.so" in x)

    fi_info_out, _ = execute_platform_cmd(fi_info)
    result.has_cxi = any(x for x in fi_info_out.split("\n") if "cxi" in x)

    return result


def is_crayex_platform() -> bool:
    """Returns True if the current platform is identified as Cray EX and
    HSTA-aware dragon package can be installed, False otherwise.

    :returns: True if current platform is Cray EX, False otherwise"""
    result = check_platform()
    return result.is_cray


@t.final
class SignalInterceptionStack(collections.abc.Collection[_TSignalHandlerFn]):
    """Registers a stack of callables to be called when a signal is
    received before calling the original signal handler.
    """

    def __init__(
        self,
        signalnum: int,
        callbacks: t.Optional[t.Iterable[_TSignalHandlerFn]] = None,
    ) -> None:
        """Set up a ``SignalInterceptionStack`` for particular signal number.

        .. note::
            This class typically should not be instanced directly as it will
            change the registered signal handler regardless of if a signal
            interception stack is already present. Instead, it is generally
            best to create or get a signal interception stack for a particular
            signal number via the `get` factory method.

        :param signalnum: The signal number to intercept
        :param callbacks: A iterable of functions to call upon receiving the signal
        """
        self._callbacks = list(callbacks) if callbacks else []
        self._original = signal.signal(signalnum, self)

    def __call__(self, signalnum: int, frame: t.Optional["FrameType"]) -> None:
        """Handle the signal on which the interception stack was registered.
        End by calling the originally registered signal hander (if present).

        :param frame: The current stack frame
        """
        for fn in self:
            fn(signalnum, frame)
        if callable(self._original):
            self._original(signalnum, frame)

    def __contains__(self, obj: object) -> bool:
        return obj in self._callbacks

    def __iter__(self) -> t.Iterator[_TSignalHandlerFn]:
        return reversed(self._callbacks)

    def __len__(self) -> int:
        return len(self._callbacks)

    @classmethod
    def get(cls, signalnum: int) -> "SignalInterceptionStack":
        """Fetch an existing ``SignalInterceptionStack`` or create a new one
        for a particular signal number.

        :param signalnum: The singal number of the signal interception stack
                          should be registered
        :returns: The existing or created signal interception stack
        """
        handler = signal.getsignal(signalnum)
        if isinstance(handler, cls):
            return handler
        return cls(signalnum, [])

    def push(self, fn: _TSignalHandlerFn) -> None:
        """Add a callback to the signal interception stack.

        :param fn: A callable to add to the unique signal stack
        """
        self._callbacks.append(fn)

    def push_unique(self, fn: _TSignalHandlerFn) -> bool:
        """Add a callback to the signal interception stack if and only if the
        callback is not already present.

        :param fn: A callable to add to the unique signal stack
        :returns: True if the callback was added, False if the callback was
                  already present
        """
        if did_push := fn not in self:
            self.push(fn)
        return did_push
